# from __future__ import annotations
from typing import List, Tuple
from abc import ABC, abstractmethod, abstractproperty

import joblib
import numpy as np
import pandas as pd
import geopandas as gpd
import scipy
from scipy.fftpack import rfft
from loguru import logger
from shapely.geometry import Point
from rasterio.crs import CRS
from skimage.transform import resize
from skimage.morphology import selem, binary_erosion, binary_dilation

from satio.rsindices import RSI_META
from satio.utils.resample import imresize, upsample, downsample_n
from satio.utils import rasterize, dem_attrs, TaskTimer
from satio.utils.geotiff import get_rasterio_profile, write_geotiff
from satio.utils.speckle import mtfilter


L2A_BANDS_10M = ['B02', 'B03', 'B04', 'B08']
L2A_BANDS_20M = ['B05', 'B06', 'B07', 'B8A', 'B11', 'B12', 'SCL',
                 'sunAzimuthAngles', 'sunZenithAngles',
                 'viewAzimuthMean', 'viewZenithMean']
L2A_BANDS_DICT = {10: L2A_BANDS_10M,
                  20: L2A_BANDS_20M}
SCL_MASK_VALUES = [0, 1, 3, 8, 9, 10, 11]

GAMMA_BANDS_DICT = {20: ['VV', 'VH']}


class Features:

    def __init__(self,
                 data: np.ndarray,
                 names: List,
                 dtype: type = np.float32,
                 attrs: dict = dict()):

        if not isinstance(names, List):
            raise TypeError("`names` should be a list.")

        if data.ndim == 2:
            data = np.expand_dims(data, axis=0)

        if data.shape[0] != len(names):
            raise ValueError('axis 0 of `data` should be equal to '
                             'the length of `names`. Instead arr.shape = '
                             f'{data.shape} and len(names) = {len(names)}.')

        if data.dtype != dtype:
            self.data = data.astype(dtype)
        else:
            self.data = data

        self.names = names
        self.attrs = attrs
        self._df = None

    def _clone(self,
               data: np.ndarray = None,
               names: List = None,
               dtype: type = None,
               attrs: dict = None):

        data = self.data if data is None else data
        names = self.names if names is None else names
        dtype = self.data.dtype if dtype is None else dtype
        attrs = self.attrs if attrs is None else attrs

        return self.__class__(data, names, dtype, attrs)

    def get_feature_index(self, feature_name):
        return self.names.index(feature_name)

    def __getitem__(self, feature_name):
        idx = self.get_feature_index(feature_name)
        return self.data[idx]

    def __setitem__(self, feature_name, value):

        if np.isscalar(value):
            value = np.ones(self.shape[-2:]) * value

        if feature_name not in self.names:
            if value.shape[-2:] != self.data.shape[-2:]:
                raise ValueError("`value` should have the same shape "
                                 "of `self.data` on the last two axis.")
            new_feat = self._clone(value, [feature_name])
            self = self.merge(new_feat)

        else:
            idx = self.get_feature_index(feature_name)
            self.data[idx] = value

    def drop(self, *features_names):
        indices = [self.get_feature_index(f) for f in features_names]
        new_data = np.delete(self.data, indices, axis=0)
        new_names = [f for f in self.names if f not in features_names]
        return self._clone(new_data, new_names)

    @property
    def df(self):
        if self._df is None:
            self._df = self._arr_to_df()
        else:
            # If needed add attrs to df
            for key, value in self.attrs.items():
                if key not in self._df.columns:
                    self._df[key] = value
        return self._df

    @property
    def shape(self):
        return self.data.shape

    @df.setter
    def df(self, value):
        if isinstance(value, pd.DataFrame):
            self._df = value
        else:
            raise TypeError("Trying to set a non DataFrame type to `df`. "
                            f"type(value) = {type(value)}")

    def _arr_to_df(self):
        features_arr, features_names = self.data, self.names
        features_df = pd.DataFrame(
            features_arr.reshape(features_arr.shape[0],
                                 features_arr.shape[1] *
                                 features_arr.shape[2]).T,
            columns=features_names)

        # Convert attributes to columns
        for key, value in self.attrs.items():
            features_df[key] = value

        return features_df

    def upsample(self, order=1) -> 'Features':
        """
        Upsamples Features to double the resolution in the space dimensions.
        Using a nearest neighbor method.
        For other resampling methods/scalings, see `imresize` method.
        """
        if order == 0:
            new_data = upsample(self.data)
        else:
            new_shape = (self.data.shape[0],
                         self.data.shape[1] * 2,
                         self.data.shape[2] * 2)

            new_data = np.empty(new_shape)
            for i in range(new_data.shape[0]):
                new_data[i, ...] = imresize(self.data[i], 2, order=order)

        return self._clone(new_data, self.names)

    def select(self, features_names, fill_value=None):

        features_names_valid = [f for f in features_names if f in self.names]
        features_names_invalid = [f for f in features_names
                                  if f not in self.names]

        if len(features_names_valid) == 0:

            if fill_value is not None:
                logger.warning("No valid features selected, returning"
                               f"constant features of value: {fill_value}")
                new_names = features_names
                new_shape = (len(features_names),
                             self.data.shape[1],
                             self.data.shape[2])
                new_data = np.empty_like(new_shape) * fill_value
                return self._clone(new_data, new_names)
            else:
                raise ValueError("No valid features selected")

        new_idxs = [self.names.index(f) for f in features_names_valid]
        new_data = self.data[new_idxs]

        if len(features_names_invalid):

            if fill_value is None:
                raise ValueError(f"Features {features_names_invalid} "
                                 "are not in the data.")
            else:
                logger.warning(f"Features {features_names_invalid} not in "
                               "data, replacing with "
                               f"constant features of value: {fill_value}")
                invalid_idxs = [features_names.index(f) for f in
                                features_names_invalid]
                for idx in invalid_idxs:
                    new_data = np.insert(new_data, idx, fill_value, axis=0)

        return self._clone(new_data, features_names)

    def merge(self, *others):
        """
        Returns a new instance merging data and names of current feature with
        multiple features.
        """
        new_names = self.names.copy()
        new_data = [self.data]

        for other in others:
            new_data.append(other.data)
            new_names.extend(other.names)

        new_data = np.concatenate(
            new_data,
            axis=0)

        return self._clone(new_data, new_names)

    def add(self,
            data: np.array,
            names: List):
        """
        Add an array with 1 or more features to the features stack
        """
        return self.merge(self.__class__(data, names))

    def add_constant(self,
                     constant_value,
                     constant_name):
        """
        Add a constant feature to the features array
        """
        if type(constant_value) is str:
            raise ValueError('`constant_value` should be numeric!')
        new_feat_arr = np.ones(self.shape[-2:]) * constant_value

        return self.merge(
            self._clone(new_feat_arr,
                        [constant_name]))

    def add_attribute(self,
                      attr_value,
                      attr_name):
        """
        Add an attribute to the features
        """
        new_attrs = self.attrs.copy()

        new_attrs[attr_name] = attr_value

        return self._clone(attrs=new_attrs)

    def add_geodataframe(self,
                         gdf,
                         feature_name,
                         bounds,
                         epsg,
                         resolution=10,
                         value_column='Index'
                         ):

        return self.merge(
            self.__class__.from_geodataframe(gdf,
                                             feature_name,
                                             bounds,
                                             epsg,
                                             resolution=resolution,
                                             value_column=value_column
                                             ))

    def add_l2a(self,
                l2a_collection,
                l2a_settings):

        return self.merge(
            self.__class__.from_l2a(l2a_collection,
                                    l2a_settings))

    def add_gamma0(self,
                   gamma0_collection,
                   gamma0_settings):
        return self.merge(
            self.__class__.from_l2a(gamma0_collection,
                                    gamma0_settings))

    def add_dem(self,
                dem_collection,
                dem_settings,
                resolution=10):

        return self.merge(
            self.__class__.from_dem(dem_collection,
                                    dem_settings,
                                    resolution=10))

    def add_latlon(self,
                   bounds,
                   epsg,
                   resolution=10):
        """
        Adds a lat, lon feature from the given bounds/epsg.

        See `from_latlon` for more details.

        """

        return self.merge(
            self.__class__.from_latlon(bounds,
                                       epsg,
                                       resolution=resolution))

    def add_pixelids(self):
        """
        Add a 'pixelids' with a different value for each pixel
        """
        new_shape = self.shape[-2:]
        new_feat_arr = np.arange(np.prod(new_shape)).reshape(new_shape)

        return self.merge(
            self._clone(new_feat_arr,
                        ['pixelids']))

    def onehot_encode(self, feature_name, prefix=None):
        """
        Returns an instance with the given the feature corresponding to
        `feature_name` onehot encoded
        """
        features = self

        if prefix is None:
            prefix = feature_name

        feat_arr = features[feature_name]
        new_features = features.drop(feature_name)

        values = np.unique(feat_arr)
        n = values.size

        encoded = np.zeros((n, *feat_arr.shape))
        encoded_names = []

        for i, v in enumerate(values):
            tmp = encoded[i]
            tmp[np.where(feat_arr == v)] = 1
            encoded[i] = tmp

            if np.isscalar(v):
                v = int(v)
            encoded_names.append(f"{prefix}_{v}")

        new_features = new_features.merge(Features(encoded, encoded_names))

        return new_features

    def show(self):
        from satio.utils.imviewer import ImViewer
        ImViewer(self.data, titles=self.names)

    def to_geotiff(self, bounds, epsg, filename):
        profile = get_rasterio_profile(self.data, bounds, epsg)
        logger.debug(f"Saving {filename}...")
        write_geotiff(self.data, profile, filename, self.names)

    @classmethod
    def from_features(cls, *features):
        return features[0].merge(*features[1:])

    @classmethod
    def from_constant(cls, value, name, shape):
        """
        Return an instance with a constant value.
        """
        data = np.ones(shape[-2:]) * value
        return cls(data, [name])

    @classmethod
    def from_geodataframe(cls,
                          gdf: gpd.GeoDataFrame,
                          feature_name,
                          bounds,
                          epsg,
                          resolution=10,
                          value_column='Index'):
        """
        Returns an instance from the rasterization of the `gdf`.
        `value_column` specifies the value for the geometries intersecting
        bounds and epsg given.
        The CRS of the `gdf` should be defined but doesn't need to be the same
        of epsg
        """
        geom_arr = rasterize(gdf,
                             bounds,
                             epsg,
                             resolution=resolution,
                             value_column=value_column)

        return cls(geom_arr, [feature_name])

    @classmethod
    def from_l2a(cls,
                 l2a_collection,
                 l2a_settings,
                 rsi_meta=None,
                 features_meta=None):

        return (l2a_collection
                .features_processor(l2a_settings,
                                    rsi_meta=rsi_meta,
                                    features_meta=features_meta)
                .compute_features())

    @classmethod
    def from_gamma0(cls,
                    gamma0_collection,
                    gamma0_settings,
                    rsi_meta=None,
                    features_meta=None):

        return (gamma0_collection
                .features_processor(gamma0_settings,
                                    rsi_meta=rsi_meta,
                                    features_meta=features_meta)
                .compute_features().upsample())

    @classmethod
    def from_dem(cls,
                 dem_collection,
                 dem_settings=None,
                 resolution=10):

        default_names = ['DEM-alt-20m', 'DEM-slo-20m',
                         'DEM-nor-20m', 'DEM-eas-20m']

        dem_settings = dem_settings or {}
        features_names = dem_settings.get('features_names',
                                          default_names)

        altitude = dem_collection.load().astype(np.float32)

        slope, aspect = dem_attrs(altitude)
        aspect = np.deg2rad(aspect)
        northness = np.cos(aspect)
        eastness = np.sin(aspect)

        default_arrs = [altitude, slope, northness, eastness]

        dem_features_dict = {fn: fa for fn, fa in zip(default_names,
                                                      default_arrs)}

        dem_features_arr = np.array(
            [dem_features_dict[fn] for fn in features_names])

        if resolution == 10:
            dem_features_arr = upsample(dem_features_arr)
        elif resolution == 20:
            pass
        else:
            raise ValueError("`resolution` should be 10 or 20.")

        return cls(dem_features_arr,
                   features_names)

    @classmethod
    def from_latlon(cls,
                    bounds,
                    epsg,
                    resolution=10):
        """
        Returns a lat, lon feature from the given bounds/epsg.

        This provide a coarse (but relatively fast) approximation to generate
        lat lon layers for each pixel. It transforms the 4 corners coordinates
        from epsg to lat lon and then extrapolate with a bilinear method
        the lat lon values of the inner pixels.
        Makes the assumption that pixels spacing is linear (which is not true)
        in lat lon, however the deviations from the true values
        (for small areas) can be neglected for the classification purposes
        in favor of the execution speed.
        """
        corners = [Point(bounds[i], bounds[j])
                   for i in [0, 2] for j in [1, 3]]
        gs = gpd.GeoSeries(corners, crs=CRS.from_epsg(epsg))
        gs_latlon = gs.to_crs(epsg=4326)

        xy = [[p.x, p.y] for p in gs_latlon.values]

        xx = np.zeros((2, 2))
        yy = np.zeros((2, 2))

        xx[1, 0] = xy[0][0]
        xx[0, 0] = xy[1][0]
        xx[1, 1] = xy[2][0]
        xx[0, 1] = xy[3][0]

        yy[1, 0] = xy[0][1]
        yy[0, 0] = xy[1][1]
        yy[1, 1] = xy[2][1]
        yy[0, 1] = xy[3][1]

        out_shape = ((bounds[3] - bounds[1]) / resolution,
                     (bounds[2] - bounds[0]) / resolution)

        lon = resize(xx, out_shape, order=1, mode='edge')
        lat = resize(yy, out_shape, order=1, mode='edge')

        features_arr = np.array([lat, lon])
        features_names = ['lat', 'lon']

        return cls(features_arr, features_names)

    def __repr__(self):
        attrs_repr = ', '.join([f"{k}: {v}" for k, v in self.attrs.items()])
        if len(attrs_repr):
            attrs_repr = " - " + attrs_repr

        return (f"<Features: {self.data.shape}{attrs_repr}>")

    def save(self, filename):
        data = [self.data,
                self.names,
                self.attrs]
        joblib.dump(data, filename)

    @classmethod
    def load(cls, filename: str) -> 'Features':
        data = joblib.load(filename)
        return cls(data[0],
                   names=data[1],
                   attrs=data[2])


class _FeaturesTimer():

    def __init__(self, *resolutions):

        self.load = {}
        self.prior_rsi = {}
        self.composite = {}
        self.interpolate = {}
        self.features = {}
        self.speckle = {}

        for r in resolutions:
            self.load[r] = TaskTimer(f'{r}m loading')
            self.prior_rsi[r] = TaskTimer(f'{r}m prior rsi calculation')
            self.composite[r] = TaskTimer(f'{r}m compositing')
            self.interpolate[r] = TaskTimer(f'{r}m interpolation')
            self.features[r] = TaskTimer(f'{r}m features computation')
            self.speckle[r] = TaskTimer(f'{r}m speckle filtering')


class BaseFeaturesProcessor(ABC):

    def __init__(self,
                 collection,
                 settings,
                 rsi_meta=None,
                 features_meta=None):

        self.collection = collection
        self.sensor = collection.sensor
        self.settings = settings
        self._bands = None
        self._rsis = None
        self._supported_bands = None
        self._supported_rsis = None

        # merge default META with provided ones
        self._rsi_meta = {**RSI_META[self.sensor], **(rsi_meta or {})}
        self._features_meta = {**FEATURES_META, **(features_meta or {})}

    @abstractproperty
    def supported_bands(self):
        ...

    @abstractproperty
    def supported_rsis(self):
        ...

    @abstractmethod
    def compute_features(self):
        ...

    @abstractproperty
    def _reflectance(self):
        """
        True for S2 and False for S1
        """
        ...

    @property
    def bands(self):
        if self._bands is None:
            self._bands = {res: [b for b in self.settings['bands']
                                 if b in sup_bands]
                           for res, sup_bands in self.supported_bands.items()}
        return self._bands

    @property
    def rsis(self):
        if self._rsis is None:
            self._rsis = {res: [b for b in self.settings['rsis']
                                if b in sup_bands]
                          for res, sup_bands in self.supported_rsis.items()}
        return self._rsis


class GAMMA0FeaturesProcessor(BaseFeaturesProcessor):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timer = _FeaturesTimer(20)

    @property
    def supported_bands(self):
        return GAMMA_BANDS_DICT

    @property
    def supported_rsis(self):
        if self._supported_rsis is None:
            rsis_dict = {}
            rsis_dict[20] = [r for r in self._rsi_meta.keys()]
            self._supported_rsis = rsis_dict

        return self._supported_rsis

    @property
    def _reflectance(self):
        return False

    def load_data(self,
                  resolution=20,
                  timeseries=None,
                  mask=None,
                  speckle_filter=True,
                  composite=True,
                  interpolate=True):
        """
        Load Timeseries from the collection and merge with `timeseries` if
        given.
        """
        from satio.timeseries import load_timeseries

        collection = self.collection
        settings = self.settings

        bands = self.bands[resolution]
        loaded_bands = timeseries.bands if timeseries is not None else []
        bands_to_load = [b for b in bands if b not in loaded_bands]

        # number of obs needs to be calculated here
        self._obs = None

        logger.info(f"GAMMA0 collection products: {collection.df.shape[0]}")

        for band in bands_to_load:

            self.timer.load[resolution].start()
            logger.info(f'{band}: loading')
            band_ts = load_timeseries(collection, band)

            # get number of observations for first band loaded
            if self._obs is None:
                self._obs = band_ts.data[0]
                self._obs = self._obs != 0
                self._obs = self._obs.sum(axis=0)

            if mask is not None:
                # mask 'nan' values. for uint16 we use 0 as nodata value
                # mask values that are False are marked as invalid
                if isinstance(mask, dict):
                    mask = mask[resolution]

                band_ts = band_ts.mask(mask)

            # drop all no data frames
            band_ts = band_ts.drop_nodata()
            self.timer.load[resolution].stop()

            if speckle_filter:
                logger.info(f"{band}: speckle filtering")
                self.timer.speckle[resolution].start()
                data = band_ts.data
                data = gamma_db_to_lin(data)
                for band_idx in range(data.shape[0]):
                    data[band_idx] = multitemporal_speckle(data[band_idx])
                band_ts.data = gamma_lin_to_db(data)
                self.timer.speckle[resolution].stop()

            composite_settings = settings.get('composite')
            if (composite_settings is not None) & composite:
                self.timer.composite[resolution].start()
                logger.info(f"{band}: compositing")
                band_ts = band_ts.composite(**composite_settings)
                self.timer.composite[resolution].stop()

            if interpolate:
                self.timer.interpolate[resolution].start()
                logger.info(f'{band}: interpolating')
                band_ts = band_ts.interpolate()
                self.timer.interpolate[resolution].stop()

            if timeseries is None:
                timeseries = band_ts
            else:
                timeseries = timeseries.merge(band_ts)

        # timeseries.data = gamma_db_to_lin(timeseries.data)
        timeseries.data = timeseries.data.astype(np.float32)
        timeseries.data /= 3000

        return timeseries

    def compute_features(self,
                         chunk_size=None,
                         upsample_order=1):

        lproc = self
        timer = self.timer
        features_meta = self._features_meta
        rsi_meta = self._rsi_meta

        if features_meta is None:
            features_meta = FEATURES_META

        # 20m processing
        resolution = 20

        ts = lproc.load_data(resolution)

        timer.features[resolution].start()
        # compute 10m features and scale to reflectance
        logger.info(f"{resolution}m: computing bands features")
        features = ts.features_from_dict(resolution,
                                         features_meta=features_meta,
                                         chunk_size=chunk_size)

        # add RSI features
        logger.info(f"{resolution}m: computing rsi features")
        rsi_names = lproc.rsis[resolution]
        features_rsi = None
        for rn in rsi_names:
            if features_rsi is None:
                features_rsi = (ts.compute_rsis(rn, rsi_meta=rsi_meta)
                                .features_from_dict(resolution,
                                                    features_meta=features_meta,  # NOQA
                                                    chunk_size=chunk_size))
            else:
                features_rsi = features_rsi.merge(
                    ts.compute_rsis(rn, rsi_meta=rsi_meta).features_from_dict(
                        resolution,
                        features_meta=features_meta,
                        chunk_size=chunk_size))

        # merge band and rsi features
        if features is None:
            features = features_rsi
        elif features_rsi is not None:
            features = features.merge(features_rsi)

        # add meta features
        features = features.merge(
            Features(np.array([self._obs]),
                     names=['gamma0_obs']))

        for r in [20]:
            timer.load[r].log()
            timer.composite[r].log()
            timer.interpolate[r].log()
            timer.features[r].log()

        return features


def multitemporal_speckle(data):
    data = mtfilter(np.rollaxis(data, 0, 3), 'gamma')
    data = np.rollaxis(data, 2, 0)
    return data


def gamma_db_to_lin(dn):
    lin = np.zeros(dn.shape, dtype=np.float32)
    lin[dn > 0] = 10 ** ((20 * np.log10(dn[dn > 0]) - 83) / 10)
    lin[lin == 0] = np.nan
    return lin


def gamma_lin_to_db(lin):
    lin[~np.isfinite(lin)] = 0
    dn = np.zeros(lin.shape)
    dn[lin > 0] = 10.0 ** ((10 * np.log10(lin[lin > 0]) + 83) / 20)
    return dn.astype(np.uint16)


class SIGMA0FeaturesProcessor(GAMMA0FeaturesProcessor):

    def load_data(self,
                  resolution=20,
                  timeseries=None,
                  mask=None,
                  speckle_filter=True,
                  composite=True,
                  interpolate=True):
        """
        Load Timeseries from the collection and merge with `timeseries` if
        given.
        """
        from satio.timeseries import load_timeseries

        collection = self.collection
        settings = self.settings

        bands = self.bands[resolution]
        loaded_bands = timeseries.bands if timeseries is not None else []
        bands_to_load = [b for b in bands if b not in loaded_bands]

        # number of obs needs to be calculated here
        self._obs = None

        logger.info(f"SIGMA0 collection products: {collection.df.shape[0]}")

        for band in bands_to_load:

            self.timer.load[resolution].start()
            logger.info(f'{band}: loading')

            # Put `mask_and_scale` to true to get real dB values
            band_ts = load_timeseries(collection, band, mask_and_scale=True)

            # Convert GEE to GAMMA, result is uint16
            band_ts.data = _gee_to_gammascaling(band_ts.data)

            # get number of observations for first band loaded
            if self._obs is None:
                self._obs = band_ts.data[0]
                self._obs = self._obs != 0
                self._obs = self._obs.sum(axis=0)

            if mask is not None:
                # mask 'nan' values. for uint16 we use 0 as nodata value
                # mask values that are False are marked as invalid
                if isinstance(mask, dict):
                    mask = mask[resolution]

                band_ts = band_ts.mask(mask)

            # drop all no data frames
            band_ts = band_ts.drop_nodata()
            self.timer.load[resolution].stop()

            if speckle_filter:
                logger.info(f"{band}: speckle filtering")
                self.timer.speckle[resolution].start()
                data = band_ts.data
                data = gamma_db_to_lin(data)
                for band_idx in range(data.shape[0]):
                    data[band_idx] = multitemporal_speckle(data[band_idx])
                band_ts.data = gamma_lin_to_db(data)
                self.timer.speckle[resolution].stop()

            composite_settings = settings.get('composite')
            if (composite_settings is not None) & composite:
                self.timer.composite[resolution].start()
                logger.info(f"{band}: compositing")
                band_ts = band_ts.composite(**composite_settings)
                self.timer.composite[resolution].stop()

            if interpolate:
                self.timer.interpolate[resolution].start()
                logger.info(f'{band}: interpolating')
                band_ts = band_ts.interpolate()
                self.timer.interpolate[resolution].stop()

            if timeseries is None:
                timeseries = band_ts
            else:
                timeseries = timeseries.merge(band_ts)

        # timeseries.data = gamma_db_to_lin(timeseries.data)
        # timeseries.data = timeseries.data.astype(np.float32)
        timeseries.data = _gammascaling_to_db(timeseries.data)

        return timeseries


def _gee_to_gammascaling(db):
    '''
    Helper function to transform GEE extracted dB data
    to GAMMA scaling in uint16
    '''
    logger.debug('Transforming GEE sigma0 to GAMMA scaled values ...')
    nodata = 0
    idxinvalid = np.where(db == nodata)
    dn = 10.0 ** ((db + 83) / 20)
    dn[idxinvalid] = nodata
    return dn.astype(np.uint16)


def _gammascaling_to_db(dn):
    '''
    Helper function to transform GAMMA scaled values
    to dB returning float
    '''
    logger.debug('Transforming GAMMA scaled values to dB...')
    nodata = 0
    idxinvalid = np.where(dn == nodata)
    db = (np.log10(dn) * 20) - 83
    db[idxinvalid] = nodata

    return db.astype(np.float)


class L2AFeaturesProcessor(BaseFeaturesProcessor):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timer = _FeaturesTimer(10, 20)

    @property
    def supported_bands(self):
        return L2A_BANDS_DICT

    @property
    def _reflectance(self):
        return True

    @property
    def supported_rsis(self):
        if self._supported_rsis is None:
            rsis_dict = {}
            rsi_res = {r: self._rsi_meta[r]['native_res']
                       for r in self._rsi_meta.keys()}
            rsis_dict[10] = [v for v, r in rsi_res.items() if r == 10]
            rsis_dict[20] = [v for v, r in rsi_res.items() if r == 20]
            self._supported_rsis = rsis_dict

        return self._supported_rsis

    def preprocess_data(self,
                        timeseries,
                        resolution,
                        mask=None,
                        composite=True,
                        interpolate=True):
        """
        Pre-processing of loaded timeseries object. Includes masking,
        compositing and interpolation.
        """
        settings = self.settings
        newtimeseries = None

        for band in timeseries.bands:
            band_ts = timeseries.select_bands([band])
            if mask is not None:
                # mask 'nan' values. for uint16 we use 0 as nodata value
                # mask values that are False are marked as invalid
                if isinstance(mask, dict):
                    mask = mask[resolution]

                band_ts = band_ts.mask(mask)

            # drop all no data frames
            band_ts = band_ts.drop_nodata()

            composite_settings = settings.get('composite')
            if (composite_settings is not None) & composite:
                self.timer.composite[resolution].start()
                logger.info(f"{band}: compositing")
                band_ts = band_ts.composite(**composite_settings)
                self.timer.composite[resolution].stop()

            if interpolate:
                self.timer.interpolate[resolution].start()
                logger.info(f'{band}: interpolating')
                band_ts = band_ts.interpolate()
                self.timer.interpolate[resolution].stop()

            if newtimeseries is None:
                newtimeseries = band_ts
            else:
                newtimeseries = newtimeseries.merge(band_ts)

        return newtimeseries

    def load_data(self,
                  resolution,
                  timeseries=None,
                  add_bands=None):
        """
        Load Timeseries from the collection and merge with `timeseries` if
        given.
        add_bands parameter gives you possibility to load additional bands.
        """
        from satio.timeseries import load_timeseries

        collection = self.collection

        bands = self.bands[resolution].copy()
        # add those bands required to calculate the requested RSI's
        rsis = self.rsis[resolution]
        for rsi in rsis:
            rsi_bands = self._rsi_meta[rsi]['bands']
            bands.extend([b for b in rsi_bands if b not in bands])
        if add_bands is not None:
            for b in add_bands:
                if b not in bands:
                    bands.extend(b)
        loaded_bands = timeseries.bands if timeseries is not None else []
        bands_to_load = [b for b in bands if b not in loaded_bands]

        if(len(bands_to_load) > 0):
            self.timer.load[resolution].start()
            bands_ts = load_timeseries(collection, *bands_to_load)
            if timeseries is None:
                timeseries = bands_ts
            else:
                timeseries = timeseries.merge(bands_ts)
            self.timer.load[resolution].stop()
        else:
            logger.info("Did not find bands to "
                        f"load for resolution: {resolution}")

        return timeseries

    def load_mask(self):
        from satio.timeseries import load_timeseries

        logger.info(f"L2A collection products: {self.collection.df.shape[0]}")

        self.timer.load[20].start()
        logger.info(f'SCL: loading')
        scl_ts = load_timeseries(self.collection, 'SCL')
        self.timer.load[20].stop()

        scl_ts = scl_ts.upsample()

        logger.info(f"SCL: preparing mask")
        mask, obs, invalid_before, invalid_after = _scl_mask(
            scl_ts.data, **self.settings['mask'])

        mask_20 = downsample_n(mask.astype(np.uint16), 1) == 1

        mask_dict = {10: mask,
                     20: mask_20}

        return mask_dict, obs, invalid_before, invalid_after

    def compute_features(self,
                         chunk_size=None,
                         upsample_order=1):

        lproc = self
        timer = self.timer
        features_meta = self._features_meta
        rsi_meta = self._rsi_meta

        if features_meta is None:
            features_meta = FEATURES_META

        mask, obs, invalid_before, invalid_after = lproc.load_mask()

        # 10m processing
        resolution = 10
        ts = None
        ts_proc = None

        # check if some 20 m indices require 10 m bands load if necessary
        add_bands = []
        rsis = lproc.rsis[20]
        sup_bands = lproc.supported_bands[resolution]
        for rsi in rsis:
            rsi_bands = self._rsi_meta[rsi]['bands']
            add_bands.extend([b for b in rsi_bands if b in sup_bands])
        ts = lproc.load_data(resolution, add_bands=add_bands)

        if ts is not None:

            # check if some rsis need to be calculated first
            rsis = lproc.rsis[resolution]
            prior_rsi = []
            post_rsi = []
            for rsi in rsis:
                if rsi_meta[rsi].get('prior', False):
                    prior_rsi.append(rsi)
                else:
                    post_rsi.append(rsi)
            if len(prior_rsi) > 0:
                logger.info(f"{resolution}m: computing prior rsis")
                timer.prior_rsi[resolution].start()
                pr_rsi = ts.compute_rsis(*prior_rsi, rsi_meta=rsi_meta)
                timer.prior_rsi[resolution].stop()

            # now pre-process the timeseries
            ts_proc = lproc.preprocess_data(ts, resolution, mask=mask)
            if len(prior_rsi) > 0:
                pr_rsi = lproc.preprocess_data(pr_rsi,  # type: ignore
                                               resolution, mask=mask)

            # start feature calculation
            timer.features[resolution].start()

            # compute 10m features and scale to reflectance
            features_10m = None
            if len(lproc.bands[resolution]) > 0:
                logger.info(f"{resolution}m: computing bands features")
                features_10m = ts_proc.select_bands(
                    lproc.bands[resolution]).features_from_dict(
                        resolution,
                        features_meta=features_meta,
                        chunk_size=chunk_size)

                # because bands features are calculated from uint16 bands
                # scale them to reflectance values
                # (fft features should not be scaled though)
                features_10m.data /= 10000

            # add RSI features
            logger.info(f"{resolution}m: computing rsi features")
            features_10m_rsi = None
            # first up, the ones already calculated:
            for rn in prior_rsi:
                if features_10m_rsi is None:
                    features_10m_rsi = pr_rsi.select_bands(  # type: ignore
                        [rn]).features_from_dict(
                        resolution,
                        features_meta=features_meta,
                        chunk_size=chunk_size)
                else:
                    features_10m_rsi = features_10m_rsi.merge(
                        pr_rsi.select_bands(  # type: ignore
                            [rn]).features_from_dict(
                            resolution,
                            features_meta=features_meta,
                            chunk_size=chunk_size))

            # now those not already calculated:
            for rn in post_rsi:
                if features_10m_rsi is None:
                    features_10m_rsi = (ts_proc
                                        .compute_rsis(rn,
                                                      rsi_meta=rsi_meta,
                                                      bands_scaling=10000)
                                        .features_from_dict(
                                            resolution,
                                            features_meta=features_meta,
                                            chunk_size=chunk_size))
                else:
                    features_10m_rsi = features_10m_rsi.merge(
                        ts_proc
                        .compute_rsis(rn,
                                      rsi_meta=rsi_meta,
                                      bands_scaling=10000)
                        .features_from_dict(resolution,
                                            features_meta=features_meta,
                                            chunk_size=chunk_size))

            # merge with band features
            if features_10m is not None:
                if features_10m_rsi is not None:
                    features_10m = features_10m.merge(features_10m_rsi)
            else:
                features_10m = features_10m_rsi

            timer.features[resolution].stop()
        else:
            features_10m = None

        # start 20 m processing
        resolution = 20
        rsis = lproc.rsis[resolution]
        if (len(lproc.bands[20]) > 0) or (len(rsis) > 0):

            features_20m = self.compute_features_20m(ts, ts_proc,
                                                     chunk_size,
                                                     mask)

            # upsample 20m features to 10m and merge them
            if features_10m is None:
                features_10m = features_20m.upsample(order=upsample_order)
            elif features_20m is not None:
                features_10m = features_10m.merge(
                    features_20m.upsample(order=upsample_order))

        # add meta features
        features_10m = features_10m.merge(
            Features(np.array([obs, invalid_before, invalid_after]),
                     names=['l2a_observations',
                            'l2a_invalid_before',
                            'l2a_invalid_after'])
        )

        for r in [10, 20]:
            timer.load[r].log()
            timer.prior_rsi[r].log()
            timer.composite[r].log()
            timer.interpolate[r].log()
            timer.features[r].log()

        return features_10m

    def compute_features_20m(self, ts, ts_proc_10m, chunk_size, mask):
        # 20m processing
        resolution = 20
        lproc = self
        timer = self.timer
        features_meta = self._features_meta
        rsi_meta = self._rsi_meta
        if features_meta is None:
            features_meta = FEATURES_META
        # load raw data
        if ts is not None:
            # downsample 10 m bands to 20m
            ts = ts.downsample()
            ts_proc_10m = ts_proc_10m.downsample()
            # load raw 20m bands
            ts = lproc.load_data(resolution, timeseries=ts)
        else:
            ts = lproc.load_data(resolution)
        # check if some rsis need to be calculated first
        rsis = lproc.rsis[resolution]
        prior_rsi = []
        post_rsi = []

        for rsi in rsis:
            if rsi_meta[rsi].get('prior', False):
                prior_rsi.append(rsi)
            else:
                post_rsi.append(rsi)

        if len(prior_rsi) > 0:
            logger.info(f"{resolution}m: computing prior rsis")
            timer.prior_rsi[resolution].start()
            pr_rsi = ts.compute_rsis(*prior_rsi, rsi_meta=rsi_meta)
            timer.prior_rsi[resolution].stop()

        # now pre-process the timeseries
        # (only those which were not pre-processed at 10 m resolution)
        bands = lproc.bands[resolution].copy()
        # add those bands required to calculate the requested RSI's
        for rsi in post_rsi:
            rsi_bands = lproc._rsi_meta[rsi]['bands']
            bands.extend([b for b in rsi_bands if b not in bands])
            bands = [b for b in bands if b not in ts_proc_10m.bands]
        ts_proc_20m = lproc.preprocess_data(ts.select_bands(bands),
                                            resolution,
                                            mask=mask)

        if ts_proc_10m is not None:
            ts_proc = ts_proc_20m.merge(ts_proc_10m)
        else:
            ts_proc = ts_proc_20m
        if len(prior_rsi) > 0:
            pr_rsi = lproc.preprocess_data(pr_rsi,  # type: ignore
                                           resolution,
                                           mask=mask)
        timer.features[resolution].start()
        logger.info(f"{resolution}m: computing rsis features")
        features_20m = None

        # compute features of prior rsis first
        for rn in prior_rsi:
            if features_20m is None:
                features_20m = pr_rsi.select_bands(  # type: ignore
                    [rn]).features_from_dict(
                    resolution,
                    features_meta=features_meta,
                    chunk_size=chunk_size)
            else:
                features_20m = features_20m.merge(
                    pr_rsi.select_bands(  # type: ignore
                        [rn]).features_from_dict(
                        resolution,
                        features_meta=features_meta,
                        chunk_size=chunk_size))
        # compute rsis features before dropping 10m bands from timeseries
        for rn in post_rsi:
            if features_20m is None:
                features_20m = (ts_proc
                                .compute_rsis(rn, rsi_meta=rsi_meta)
                                .features_from_dict(resolution,
                                                    features_meta=features_meta,  # NOQA
                                                    chunk_size=chunk_size))
            else:
                features_20m = features_20m.merge(
                    ts_proc
                    .compute_rsis(rn,
                                  rsi_meta=rsi_meta,
                                  bands_scaling=10000)
                    .features_from_dict(resolution,
                                        features_meta=features_meta,
                                        chunk_size=chunk_size))

        if len(lproc.bands[resolution]) > 0:
            # reduce timeseries to 20m bands only
            ts_proc = ts_proc.select_bands(lproc.bands[resolution])

            # compute bands features
            logger.info(f"{resolution}m: computing bands features")
            features_20m_bands = ts_proc.features_from_dict(
                resolution,
                features_meta=features_meta,
                chunk_size=chunk_size)

            # scale bands features  (fft features should not be scaled though)
            features_20m_bands.data /= 10000

            # merge with other features
            if features_20m is None:
                features_20m = features_20m_bands
            else:
                features_20m = features_20m.merge(features_20m_bands)

        timer.features[resolution].stop()

        return features_20m


class MAJAFeaturesProcessor(L2AFeaturesProcessor):

    def load_mask(self):
        from satio.timeseries import load_timeseries

        logger.info(f"L2A collection products: {self.collection.df.shape[0]}")

        self.timer.load[10].start()
        logger.info(f'CLM: loading')
        scl_ts = load_timeseries(self.collection, 'CLM')
        self.timer.load[10].stop()

        mask = scl_ts.data[0] == 0
        obs = np.ones(mask.shape[-2:]) * mask.shape[0]
        invalid_before = 1 - mask.mean(axis=0)
        invalid_after = invalid_before

        mask_20 = downsample_n(mask.astype(np.uint16), 1) == 1

        mask_dict = {10: mask,
                     20: mask_20}

        return mask_dict, obs, invalid_before, invalid_after

    def load_data(self,
                  resolution,
                  timeseries=None,
                  mask=None,
                  composite=True,
                  interpolate=True):
        """
        Load Timeseries from the collection and merge with `timeseries` if
        given.
        """
        from satio.timeseries import load_timeseries

        collection = self.collection
        settings = self.settings

        bands = self.bands[resolution]
        loaded_bands = timeseries.bands if timeseries is not None else []
        bands_to_load = [b for b in bands if b not in loaded_bands]

        for band in bands_to_load:

            self.timer.load[resolution].start()
            logger.info(f'{band}: loading')
            band_ts = load_timeseries(collection, band)

            # if (band == 'B08') and (resolution == 20):
            #     band_ts = band_ts.downsample()

            if mask is not None:
                # mask 'nan' values. for uint16 we use 0 as nodata value
                # mask values that are False are marked as invalid
                if isinstance(mask, dict):
                    mask = mask[resolution]

                band_ts = band_ts.mask(mask, nodata_value=-10000)

            band_ts.data[band_ts.data == -10000] = 0
            band_ts.data = band_ts.data.astype(np.uint16)

            # drop all no data frames
            band_ts = band_ts.drop_nodata()
            self.timer.load[resolution].stop()

            composite_settings = settings.get('composite')
            if (composite_settings is not None) & composite:
                self.timer.composite[resolution].start()
                logger.info(f"{band}: compositing")
                band_ts = band_ts.composite(**composite_settings)
                self.timer.composite[resolution].stop()

            if interpolate:
                self.timer.interpolate[resolution].start()
                logger.info(f'{band}: interpolating')
                band_ts = band_ts.interpolate()
                self.timer.interpolate[resolution].stop()

            if timeseries is None:
                timeseries = band_ts
            else:
                timeseries = timeseries.merge(band_ts)

        return timeseries


class ICORFeaturesProcessor(L2AFeaturesProcessor):

    def load_mask(self):
        from satio.timeseries import load_timeseries

        logger.info(f"L2A collection products: {self.collection.df.shape[0]}")

        self.timer.load[10].start()
        logger.info(f'IPX: loading')
        scl_ts = load_timeseries(self.collection, 'IPX')
        self.timer.load[10].stop()

        mask = scl_ts.data[0] == 1

        mask_10 = upsample(mask.astype(np.uint8)) == 1
        obs = np.ones(mask_10.shape[-2:]) * mask_10.shape[0]
        invalid_before = 1 - mask_10.mean(axis=0)
        invalid_after = invalid_before

        mask_dict = {10: mask_10,
                     20: mask}

        # zero_arr = np.zeros(mask_10.shape[-2:])

        return mask_dict, obs, invalid_before, invalid_after

    def load_data(self,
                  resolution,
                  timeseries=None,
                  mask=None,
                  composite=True,
                  interpolate=True):
        """
        Load Timeseries from the collection and merge with `timeseries` if
        given.
        """
        from satio.timeseries import load_timeseries

        collection = self.collection
        settings = self.settings

        bands = self.bands[resolution]
        loaded_bands = timeseries.bands if timeseries is not None else []
        bands_to_load = [b for b in bands if b not in loaded_bands]

        for band in bands_to_load:

            self.timer.load[resolution].start()
            logger.info(f'{band}: loading')
            band_ts = load_timeseries(collection, band)

            # if (band == 'B08') and (resolution == 20):
            #     band_ts = band_ts.downsample()

            if mask is not None:
                # mask 'nan' values. for uint16 we use 0 as nodata value
                # mask values that are False are marked as invalid
                if isinstance(mask, dict):
                    mask = mask[resolution]

                band_ts = band_ts.mask(mask, nodata_value=0)

            band_ts.data[band_ts.data < 0] = 0
            band_ts.data = band_ts.data.astype(np.uint16)

            # drop all no data frames
            band_ts = band_ts.drop_nodata()
            self.timer.load[resolution].stop()

            composite_settings = settings.get('composite')
            if (composite_settings is not None) & composite:
                self.timer.composite[resolution].start()
                logger.info(f"{band}: compositing")
                band_ts = band_ts.composite(**composite_settings)
                self.timer.composite[resolution].stop()

            if interpolate:
                self.timer.interpolate[resolution].start()
                logger.info(f'{band}: interpolating')
                band_ts = band_ts.interpolate()
                self.timer.interpolate[resolution].stop()

            if timeseries is None:
                timeseries = band_ts
            else:
                timeseries = timeseries.merge(band_ts)

        return timeseries


def get_array_windows(n_rows, n_cols, chunk_size):
    """
    Return tuple of tuples ((row_start, row_end), (col_start, col_end))
    for a set of windows of arrays of shape 'chunk_size'.
    """
    if isinstance(chunk_size, int):
        chunk_size = (chunk_size, chunk_size)

    rows_ids = list(range(0, n_rows, chunk_size[0])) + [n_rows]
    cols_ids = list(range(0, n_cols, chunk_size[1])) + [n_cols]

    windows = []
    for ridx in range(len(rows_ids) - 1):
        for cidx in range(len(cols_ids) - 1):
            win = ((rows_ids[ridx], rows_ids[ridx + 1]),
                   (cols_ids[cidx], cols_ids[cidx + 1]))
            windows.append(win)

    return windows


def compute_chunked_features(data,
                             func,
                             n_features,
                             chunk_size=256,
                             **kwargs):
    raise NotImplementedError


def _scl_mask(scl_data,
              *,
              mask_values,
              erode_r=None,
              dilate_r=None,
              max_invalid_ratio=None):
    """
    From a timeseries (t, y, x) returns a binary mask False for the
    given mask_values and True elsewhere.

    Parameters:
    -----------
    slc_data: 3D array
        Input array for computing the mask

    mask_values: list
        values to set to False in the mask

    erode_r : int
        Radius for eroding disk on the mask

    dilate_r : int
        Radius for dilating disk on the mask

    max_invalid_ratio : float
        Will set mask values to True, when they have an
        invalid_ratio > max_invalid_ratio

    Returns:
    --------
    mask : 3D bool array
        mask True for valid pixels, False for invalid

    obs : 2D int array
        number of valid observations (different from 0 in scl_data)

    invalid_before : 2D float array
        ratio of invalid obs before morphological operations

    invalid_after : 2D float array
        ratio of invalid obs after morphological operations
    """
    scl_data = np.squeeze(scl_data)

    ts_obs = scl_data != 0

    obs = ts_obs.sum(axis=0)

    mask = np.isin(scl_data, mask_values)
    ma_mask = (mask & ts_obs)

    invalid_before = ma_mask.sum(axis=0) / obs

    if (erode_r is not None) | (erode_r > 0):
        erode_disk = selem.disk(erode_r)
        for i in range(mask.shape[0]):
            mask[i] = binary_erosion(mask[i], erode_disk)

    if (dilate_r is not None) | (dilate_r > 0):
        dilate_disk = selem.disk(dilate_r)
        for i in range(mask.shape[0]):
            mask[i] = binary_dilation(mask[i], dilate_disk)

    ma_mask = (mask & ts_obs)
    invalid_after = ma_mask.sum(axis=0) / obs

    # invert values to have True for valid pixels and False for clouds
    mask = ~mask

    if max_invalid_ratio is not None:
        max_invalid_mask = invalid_after > max_invalid_ratio
        mask = mask | np.broadcast_to(max_invalid_mask, mask.shape)

    return mask, obs, invalid_before, invalid_after


class _iLocIndexer:

    def __init__(self, feat_coll):
        self._feat_coll = feat_coll

    def __getitem__(self, idx):
        key = self._feat_coll.indices[idx]
        return self._feat_coll.__getitem__(key)


class FeaturesCollection:

    def __init__(self, features: List[Features], indices: List = None):

        if indices is None:
            indices = list(range(len(features)))

        def _raise_type(f, i):
            raise TypeError(f"'features' should be Features objects. "
                            f'Instead is {type(f)} at index {i}.')

        self.features = {i: f.add_attribute(i, 'collection_idx')
                         if isinstance(f, Features)
                         else _raise_type(f, i)
                         for f, i in zip(features, indices)}

        self.indices = list(indices)

        self._df = None
        self._iter_idx = 0

    def __getitem__(self, key):

        if isinstance(key, list):
            return [self.features[k] for k in key]
        else:
            return self.features[key]

    def __setitem__(self, key, value):
        if not isinstance(value, Features):
            raise TypeError('value type should be `Features`')
        self.features[key] = value
        self.indices.extend(key)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    @property
    def df(self):
        if self._df is None:
            self._df = pd.concat([f.df for f in self.features.values()],
                                 ignore_index=True, axis=0)
        return self._df

    @classmethod
    def from_df(cls,
                df: pd.DataFrame,
                attr_cols: List[str] = [],
                index_column: str = None,
                canvas_shape: Tuple[int] = None):
        """
        index_column should be the column name of a unique location identifier.
        The canvas shape is assumed to be square if None is provided.

        If no index_column is provided, the indices will be read from the
        default column 'collection_idx' which is created when generating
        a DataFrame from a FeaturesCollection
        """
        if index_column is None:
            index_column = 'collection_idx'

        indices = df[index_column].unique().tolist()

        cols_drop = attr_cols + [index_column]

        features = []
        features_names = [c for c in df.columns
                          if c not in cols_drop]

        for idx in indices:
            feat_df = df[df[index_column] == idx]

            if canvas_shape is None:
                rows = cols = int(feat_df.shape[0] ** 0.5)
                canvas_shape = rows, cols

            features_arr = (feat_df
                            .drop(columns=cols_drop)
                            .values
                            .T
                            .reshape(len(features_names),
                                     canvas_shape[0],
                                     canvas_shape[1])
                            )

            features.append(Features(features_arr,
                                     features_names,
                                     attrs={k: feat_df[k].unique()[0]
                                            for k in attr_cols}))

        return cls(features, indices)

    def __repr__(self):
        return f'<FeaturesCollection - n: ({len(list(self.features.keys()))})>'

    def __iter__(self):
        self._iter_idx = 0
        return self

    def __next__(self):
        try:
            idx = self.indices[self._iter_idx]
        except IndexError:
            raise StopIteration
        self._iter_idx += 1
        return self[idx]


"""
Below is a set of deafault temporal features computation functions
"""


def fft_coeffs(x, n_coeffs=5):
    t = rfft(x, axis=0)
    t = t[:n_coeffs, ...]
    return t


def tsteps(x, n_steps=6):
    return scipy.signal.resample(x, n_steps, axis=0)


def percentile(x, q=[10, 30, 50, 70, 90]):
    return np.percentile(x, q=q, axis=0)


def percentile_iqr(x, q=[10, 50, 90], iqr=[25, 75]):

    q_all = list(set(q + iqr))
    q_ids = [q_all.index(qv) for qv in q]
    iqr_ids = [q_all.index(qv) for qv in iqr]

    perc = np.percentile(x, q=q_all, axis=0)

    perc_arr = perc[q_ids]
    iqr_arr = perc[iqr_ids[1]] - perc[iqr_ids[0]]

    iqr_arr = np.expand_dims(iqr_arr, axis=0)

    arr = np.concatenate([perc_arr, iqr_arr], axis=0)

    return arr


def std(x):
    return np.std(x, axis=0)


FEATURES_META = {
    "percentile_iqr": {
        "function": percentile_iqr,
        "parameters": {
            'q': [10, 50, 90],
            'iqr': [25, 75]
        },
        "names": ['p10', 'p50', 'p90', 'iqr']
    },
    "tsteps": {
        "function": tsteps,
        "parameters": {
            'n_steps': 6,
        },
        "names": ['ts0', 'ts1', 'ts2', 'ts3', 'ts4', 'ts5']
    }
}

# features_meta = {
#   "percentile": {
#       "function": percentile,
#       "parameters": {
#           'q': [10, 30, 50, 70, 90],
#       },
#       "names": ['p10', 'p30', 'p50', 'p70', 'p90']
#   },
#   "std": {
#       "function": std,
#       "names": ['std']
#   },
#   "fft_coeffs": {
#       "function": fft_coeffs,
#       "parameters": {
#           'n_coeffs': 5,
#       },
#       "names": ['fft00', 'fft01', 'fft02', 'fft03', 'fft04']
#   },
#   "tsteps": {
#       "function": tsteps,
#       "parameters": {
#           'n_steps': 6,
#       },
#       "names": ['ts0', 'ts1', 'ts2', 'ts3', 'ts4', 'ts5']
#   }
# }
