import os
import json
import random
from pathlib import Path
import xml.etree.ElementTree as ET
from typing import List
import concurrent.futures

from loguru import logger
from tqdm.auto import tqdm
import boto3
import numpy as np
import xarray as xr
import geopandas as gpd
from shapely.geometry import Polygon
import rasterio
from rasterio.crs import CRS
from rasterio.warp import reproject
from rasterio.errors import RasterioIOError
from rasterio.session import AWSSession
from skimage.transform import rescale


S2_BANDS = 'B01,B02,B03,B04,B05,B06,B07,B08,B8A,B09,B10,B11,B12'.split(',')

L2A_BANDS = S2_BANDS + ['SCL']

L2A_HRES_BANDS = ['B02', 'B03', 'B04', 'B05',
                  'B06', 'B07', 'B08', 'B8A',
                  'B11', 'B12', 'SCL']

L1C_BANDS = S2_BANDS
GRD_BANDS = ['VV', 'VH']
FEATURE_TYPES = ['L2A', 'L1C', 'GRD-DESCENDING', 'GRD-ASCENDING', 'GOOGLEMAPS']

L2A_10M_BANDS = ['AOT', 'B02', 'B03', 'B04', 'B08', 'TCI', 'WVP']
L2A_20M_BANDS = ['AOT', 'B02', 'B03', 'B04', 'B05',
                 'B06', 'B07', 'B08', 'B11', 'B12',
                 'B8A', 'SCL', 'TCI', 'WVP']
L2A_60M_BANDS = ['AOT', 'B01', 'B02', 'B03', 'B04',
                 'B05', 'B06', 'B07', 'B08', 'B09',
                 'B10', 'B11', 'B12', 'B8A', 'SCL', 'TCI',
                 'WVP']

L2A_RES_BANDS = [L2A_10M_BANDS,
                 L2A_20M_BANDS,
                 L2A_60M_BANDS]


class ParallelLoader:

    def __init__(self,
                 max_workers=20,
                 progressbar=False,
                 rio_gdal_options=None,
                 fill_value=0,
                 random_order=True):

        self._max_workers = max_workers
        self._progressbar = progressbar

        if rio_gdal_options is None:
            rio_gdal_options = {'GDAL_CACHEMAX': 0}

        self._rio_gdal_options = rio_gdal_options
        self._fill_value = fill_value
        self._random_order = random_order

    def _load_array_bounds(self, fname, bounds):

        with rasterio.Env(**self._rio_gdal_options):
            with rasterio.open(fname) as src:
                window = rasterio.windows.from_bounds(*bounds, src.transform)

                vals = np.array(window.flatten())
                if (vals % 1 > 0).any():

                    # logger.warning("Rounding floating window"
                    #                "offsets and shape")
                    window = window.round_lengths(op='ceil').round_offsets()

                arr = src.read(window=window, boundless=True,
                               fill_value=self._fill_value)

        arr = np.squeeze(arr)
        return arr

    def _load_array_bounds_safe(self, fname, bounds, max_retries=50):
        """
        to be tested
        """
        with rasterio.Env(**self._rio_gdal_options):
            with rasterio.open(fname) as src:
                window = rasterio.windows.from_bounds(*bounds, src.transform)

                try:
                    arr = src.read(window=window, boundless=True,
                                   fill_value=self._fill_value)
                    arr = np.squeeze(arr)

                except RasterioIOError as e:
                    if max_retries:
                        logger.error(f"Error reading data from {fname}: {e}"
                                     " Retrying")
                        return self._load_array_bounds_safe(fname,
                                                            bounds,
                                                            max_retries - 1)
                    else:
                        logger.error(f"Error reading data from {fname}: {e}"
                                     " Returning an array of zeros.")
                        arr = np.zeros((window.height, window.width))

        return arr

    def load_arrays(self, filenames, bounds):

        def f_handle(filename):
            return self._load_array_bounds(filename, bounds)

        if self._random_order:
            ids = list(range(len(filenames)))
            ids_random = ids.copy()
            random.shuffle(ids_random)
            ids_map = [ids_random.index(i) for i in ids]
            filenames = [filenames[i] for i in ids_random]

        arrs_list = self._run_parallel(f_handle,
                                       filenames,
                                       self._max_workers,
                                       threads=True,
                                       progressbar=self._progressbar)

        if self._random_order:
            arrs_list = [arrs_list[i] for i in ids_map]

        return arrs_list

    def _load_raw(self, collection, bands, resolution=10):

        bounds = collection.bounds

        filenames = [f for band in bands
                     for f in collection.get_band_filenames(band, resolution)]

        arrs_list_flat = self.load_arrays(filenames, bounds)

        arrs_list = self._split_arrs_list(arrs_list_flat,
                                          n_bands=len(bands))

        return arrs_list

    def _split_arrs_list(self,
                         arrs: List[np.ndarray],
                         n_bands: int) -> List[np.ndarray]:
        """
        'arrs' is a list of arrays with shape [y, x] loaded from filenames.
        The length of the list should be n_bands * n_timestamps.
        This function returns a list of leght n_bands with arrays of shape
        [n_timestamps, y, x].
        """
        n_timestamps = len(arrs) // n_bands
        if len(arrs) % n_bands:
            raise ValueError("Number of files loaded is not a multiple "
                             "of the number of bands as expected.")

        out_list = [np.array(arrs[n_timestamps * i:n_timestamps * (i + 1)])
                    for i in range(n_bands)]

        return out_list

    def load(self, collection, bands, resolution, resample=False):
        """
        Load data for the given bands. resolution should be 10 or 20.
        If resolution is 10, only 'B02', 'B03', 'B04' and 'B08' will be
        loaded at 10 m resolution, the rest at 20 m. Unless 'resample' is
        set to True. In that case, 20 m bands will be upsampled to 10 m with
        a bilinear filter (except 'SCL' which will be resampled with
        nearest neighbors method).
        """
        if not isinstance(bands, (list, tuple)):
            raise TypeError("'bands' should be a list/tuple of bands. "
                            f"Its type is: {type(bands)}")
        if resolution not in [10, 20, 60]:
            raise ValueError(f"Resolution value: '{resolution}' not supported"
                             " Should be 10, 20 or 60.")

        arrs_list = self._load_raw(collection, bands, resolution)

        if resample:
            arrs_list = self._resample(arrs_list, bands, resolution, order=1)

        products = collection.products
        timestamps = collection.timestamps
        bands = list(bands)
        bounds = list(collection.bounds)
        epsg = collection.epsg

        xds_dict = {band: self._arr_to_xarr(arrs_list[i],
                                            bounds,
                                            timestamps,
                                            name=band)
                    for i, band in enumerate(bands)}

        xds_dict.update({'epsg': epsg,
                         'bounds': bounds,
                         'products': products,
                         'bands': bands})
        # xds = xr.Dataset(xds_dict)

        return xds_dict

    @staticmethod
    def _run_parallel(f, my_iter, max_workers, threads=True, progressbar=True):

        if threads:
            Pool = concurrent.futures.ThreadPoolExecutor
        else:
            Pool = concurrent.futures.ProcessPoolExecutor

        with Pool(max_workers=max_workers) as executor:
            if progressbar:
                results = list(
                    tqdm(executor.map(f, my_iter), total=len(my_iter)))
            else:
                results = list(executor.map(f, my_iter))

        return results

    def _resample(self, arrs, bands, resolution, order=1):

        if resolution == 20:
            return arrs

        elif resolution == 10:
            new_arrs = []
            for i, b in enumerate(bands):

                arr = arrs[i]

                if b not in L2A_10M_BANDS:
                    scale = 2
                    order = 0 if b == 'SCL' else order
                    arr = self._resample_arr(arr, scale, order)

                new_arrs.append(arr)
        else:
            raise NotImplementedError("Can only resample 10 m or 20 m bands"
                                      "resolution.")

        return new_arrs

    @staticmethod
    def _resample_arr(arr, scale, order):
        arr = np.transpose(arr, [1, 2, 0])
        arr = rescale(arr,
                      scale=scale,
                      order=order,
                      preserve_range=True,
                      multichannel=True,
                      anti_aliasing=True)

        arr = np.transpose(arr, [2, 0, 1])

        return arr

    @staticmethod
    def _arr_resolution(arr,
                        bounds):

        x_meters = bounds[2] - bounds[0]
        return x_meters // arr.shape[-1]

    def _arr_to_xarr(self,
                     arr,
                     bounds,
                     timestamps,
                     name=None):

        if arr.ndim == 1:
            # single pixel
            arr = np.expand_dims(arr, axis=-1)
            arr = np.expand_dims(arr, axis=-1)

        resolution = self._arr_resolution(arr, bounds)

        dims = ['timestamp', 'y', 'x']
        dims = {k: arr.shape[i] for i, k in enumerate(dims)}

        center_shift = resolution / 2
        xmin, xmax = (bounds[0] + center_shift), (bounds[2] - center_shift)
        ymin, ymax = (bounds[1] + center_shift), (bounds[3] - center_shift)

        x = np.linspace(xmin,
                        xmax,
                        dims['x'])

        y = np.linspace(ymin,
                        ymax,
                        dims['y'])

        coords = {'timestamp': timestamps,
                  'x': x, 'y': y}

        da = xr.DataArray(arr,
                          coords=coords,
                          dims=dims,
                          name=name,
                          attrs={'resolution': resolution})

        return da


class AWSParallelLoader(ParallelLoader):

    def __init__(self,
                 aws_access_key_id,
                 aws_secret_access_key,
                 region_name='eu-central-1',
                 rio_gdal_options=None,
                 max_workers=20,
                 fill_value=0):

        boto3_session = boto3.Session(aws_access_key_id,
                                      aws_secret_access_key,
                                      region_name=region_name)
        self._session = AWSSession(boto3_session)

        if rio_gdal_options is None:
            rio_gdal_options = {
                'AWS_REQUEST_PAYER': 'requester',
                'GDAL_DISABLE_READDIR_ON_OPEN': 'EMPTY_DIR',
                'CPL_VSIL_CURL_ALLOWED_EXTENSIONS': 'jp2',
                'VSI_CACHE': False}

        super().__init__(rio_gdal_options=rio_gdal_options,
                         max_workers=max_workers,
                         fill_value=fill_value)

    def _load_array_bounds(self, fname, bounds):

        with rasterio.Env(session=self._session,
                          **self._rio_gdal_options):
            with rasterio.open(fname) as src:
                window = rasterio.windows.from_bounds(*bounds, src.transform)
                arr = src.read(window=window, boundless=True,
                               fill_value=self._fill_value)

        arr = np.squeeze(arr)
        return arr


class CorruptedL2AProduct(Exception):
    pass


def get_l2a_filenames(filename):

    filename = Path(filename)

    # check if it's a product downloaded from AWS
    aws_sub = filename / 'R10m'
    if aws_sub.is_dir():
        return get_l2a_aws_filenames(filename)

    try:
        xml_manifest = 'MTD_MSIL2A.xml'
        metadata = os.path.join(filename, xml_manifest)
        root = ET.parse(metadata).getroot()
        filenames = dict()
        if len(root[0][0][-1]) == 1:
            for i, resolution in enumerate([60, 20, 10]):
                filenames[resolution] = {child.text.split('_')[-2]: os.path.join(filename, child.text + '.jp2')  # NOQA
                                         for child in root[0][0][-1][0][0] if child.text.endswith("{}m".format(resolution))}  # NOQA
        else:
            for i, resolution in enumerate([60, 20, 10]):
                filenames[resolution] = {child.text.split('_')[-2]: os.path.join(filename, child.text + '.jp2')  # NOQA
                                         for child in root[0][0][-1][i][0]}

        return invert_band_res_keys(filenames)
    except Exception:
        raise CorruptedL2AProduct(f"Error with product: {filename}.\n"
                                  f"Failed to parse metadata file: {metadata}")


def get_l2a_aws_filenames(product_path):
    product_path = Path(product_path)

    r10_files = [product_path / 'R10m' / f'B{n:02d}.jp2'
                 for n in [2, 3, 4, 8]]
    r20_files = ([product_path / 'R20m' / f'B{n:02d}.jp2'
                  for n in [5, 6, 7, 11, 12]]
                 + [product_path / 'R20m' / 'SCL.jp2'])
    r60_files = [product_path / 'R60m' / 'SCL.jp2']

    filenames = {10: {str(f.name)[:3]: f for f in r10_files},
                 20: {str(f.name)[:3]: f for f in r20_files},
                 60: {str(f.name)[:3]: f for f in r60_files}}

    return invert_band_res_keys(filenames)


def invert_band_res_keys(filenames):
    """
    Invert first and second key in dictionary of jp2 filenames
    """
    filenames2 = dict()
    for res, bands_dict in filenames.items():
        for b, f in bands_dict.items():
            filenames2[b] = filenames2.get(b, dict())
            filenames2[b][res] = f
    return filenames2


def get_jp2_filenames(filenames, band, resolution):
    """
    Returns jp2 file for given band with native resolution closest/lower
    than specified value
    """
    available_res = np.array(list(filenames[band].keys()))
    min_avl = np.min(available_res)
    if resolution <= min_avl:
        target_res = min_avl
    else:
        diff = available_res - resolution
        target_res = resolution if 0 in diff else (
            np.max(diff[diff < 0]) + resolution)
    return filenames[band][target_res]


def get_aws_filenames(product_key: str) -> dict:
    """
    Returns a dictionary of filenames with the structure
    filenames[band][resolution] where band is one of the
    Sentinel-2 bands 'B01', 'B02', ..., 'B12' and resolution
    is one of 10, 20, 60.

    Parameters
    ----------
    product_key: str
        String pointing to the S3 path of the Sentinel-2 product.
        The function will append the conventional tree structure to
        the jp2 files and produce the filenames from this.
        e.g.
    """
    filenames = dict()
    resolutions = [10, 20, 60]
    for bands, res in zip(L2A_RES_BANDS, resolutions):
        for b in bands:
            band_filenames = filenames.get(b, dict())
            band_filenames[res] = os.path.join(
                product_key, f'R{res}m', f'{b}.jp2')
            filenames[b] = band_filenames

    return filenames


def get_s3_filenames(product_key, band, resolution):
    filenames_dict = get_aws_filenames(product_key)
    jp2_filenames = get_jp2_filenames(filenames_dict, band, resolution)
    return jp2_filenames


class NpEncoder(json.JSONEncoder):
    """
    Encode NumPy arrays to JSON
    """

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


def load_array_bounds(fname, bounds, fill_value=0):

    with rasterio.open(fname) as src:

        window = rasterio.windows.from_bounds(*bounds, src.transform)
        arr = src.read(window=window, boundless=True,
                       fill_value=fill_value)

    return np.squeeze(arr)


def load_reproject(filename, bounds, epsg, resolution=10, border_buff=2):
    """
    Read from latlon layer and reproject to UTM
    """
    bbox = gpd.GeoSeries(Polygon.from_bounds(*bounds), crs=CRS.from_epsg(epsg))

    bounds = bbox.buffer(
        border_buff * resolution).to_crs(epsg=4326).bounds.values[0]
    utm_bounds = bbox.buffer(
        border_buff * resolution).bounds.values[0].tolist()

    width = int((utm_bounds[2] - utm_bounds[0]) / resolution)
    height = int((utm_bounds[3] - utm_bounds[1]) / resolution)

    gim = load_array_bounds(filename, bounds)

    src_crs = CRS.from_epsg(4326)
    dst_crs = CRS.from_epsg(bbox.crs.to_epsg())

    src_transform = rasterio.transform.from_bounds(
        *bounds, gim.shape[1], gim.shape[0])
    dst_transform = rasterio.transform.from_bounds(*utm_bounds, width, height)

    dst = np.zeros((height, width))

    reproject(gim,
              dst,
              src_transform=src_transform,
              dst_transform=dst_transform,
              src_crs=src_crs,
              dst_crs=dst_crs)

    dst_final = dst[border_buff:-border_buff, border_buff:-border_buff]

    return dst_final
