import json
from concurrent.futures import ThreadPoolExecutor

import matplotlib.pyplot as plt
import geopandas as gpd
import numpy as np
import shapely
from shapely.geometry import Polygon
from shapely.geometry import mapping, shape
from fiona.transform import transform_geom
from pyproj.crs import CRS
import rasterio
from rasterio.windows import Window  # NOQA

import satio
from satio.layers import load_s2tile_windows


class BBoxGrid:
    """
    Divides a rectangular box in a grid of bboxes of given size in pixels
    """

    def __init__(self,
                 bounds,
                 bbox_size=256,
                 resolution=10,
                 epsg=None,
                 **kwargs):
        """
        bounds = (xmin, ymin, xmax, ymax)
        """
        bounds = self._round_bounds(bounds, resolution)
        left, bottom, right, top = bounds
        self.origin_x = left
        self.origin_y = top
        self.width = right - left
        self.height = top - bottom
        self.epsg = epsg

        # make it compatible with previous scripts, where 'resolution'
        # parameter was called 'pixelsize'
        resolution = kwargs.get('pixelsize', resolution)
        self.bbox_size = bbox_size
        self.bbox_size_m = bbox_size * resolution

        self.nrows = self._get_nrows()
        self.ncols = self._get_ncols()

        self.n_bboxes = self.nrows * self.ncols

    @staticmethod
    def _round_bounds(bounds, resolution):
        """
        Buffer bounds of 'resolution' meters and then round to the closest
        multiple of resolution. This will split in bboxes that are multiple of
        the desired pixels resolution.
        """
        bounds = np.array(bounds)
        bounds = bounds + np.array([-resolution, -resolution,
                                    resolution, resolution])
        bounds = bounds // resolution * resolution
        return bounds.tolist()

    def _get_nrows(self):
        return np.ceil(self.height / self.bbox_size_m).astype(int)

    def _get_ncols(self):
        return np.ceil(self.width / self.bbox_size_m).astype(int)

    def _get_bbox_row_col(self, box_id):
        return list(zip(*np.unravel_index([box_id], (self.nrows, self.ncols),
                                          order='C')))[0]

    def _get_bbox_bounds(self, row, col):
        left, top = (self.origin_x + col * self.bbox_size_m,
                     self.origin_y - row * self.bbox_size_m)
        right, bottom = left + self.bbox_size_m, top - self.bbox_size_m
        return left, bottom, right, top

    def __len__(self):
        return self.n_bboxes

    @property
    def shape(self):
        return self.nrows, self.ncols

    def __getitem__(self, bbox_id):

        if isinstance(bbox_id, slice):
            start = bbox_id.start if bbox_id.start else 0
            stop = bbox_id.stop if bbox_id.stop else len(self)
            step = bbox_id.step if bbox_id.step else 1

            if start < 0:
                start += len(self)

            if stop < 0:
                stop += len(self)

            return [self._get_bbox(i) for i in range(start, stop, step)]
        else:
            return self._get_bbox(bbox_id)

    def _get_bbox(self, bbox_id):
        if bbox_id < 0:
            bbox_id = len(self) + bbox_id

        if (bbox_id < 0) or (bbox_id >= (self.nrows * self.ncols)):
            raise IndexError("box_id should be between 0 and {}"
                             .format(len(self) - 1))

        row, col = self._get_bbox_row_col(bbox_id)
        left, bottom, right, top = self._get_bbox_bounds(row, col)
        return BBox((left, bottom, right, top), bbox_id)

    def __repr__(self):
        return "<BboxGrid: bbox_size={}, n_boxes={}>".format(self.bbox_size,
                                                             len(self))

    def as_gdf(self, aoi_shp=None):
        """
        Returns GeoDataFrame of bboxes in grid. If aoi_shp: Polygon is given,
        the GeoDataFrame will contain only the intersecting bboxes
        """
        if aoi_shp:
            return gpd.GeoDataFrame({'id': b.id, 'geometry': b}
                                    for b in self if b.intersects(aoi_shp))
        else:
            return gpd.GeoDataFrame({'id': b.id, 'geometry': b} for b in self)


class BBoxGridShp(BBoxGrid):
    """Use pure shapely boxes instead of custom BBox"""

    def _get_bbox(self, bbox_id):
        if bbox_id < 0:
            bbox_id = len(self) + bbox_id

        if (bbox_id < 0) or (bbox_id >= (self.nrows * self.ncols)):
            raise IndexError("box_id should be between 0 and {}"
                             .format(len(self) - 1))

        row, col = self._get_bbox_row_col(bbox_id)
        left, bottom, right, top = self._get_bbox_bounds(row, col)
        return shapely.geometry.box(left, bottom, right, top)

    def as_gdf(self, aoi_shp=None):
        """
        Returns GeoDataFrame of bboxes in grid. If aoi_shp: Polygon is given,
        the GeoDataFrame will contain only the intersecting bboxes
        """
        if aoi_shp:
            return gpd.GeoDataFrame({'id': i, 'geometry': b}
                                    for i, b in enumerate(self)
                                    if b.intersects(aoi_shp))
        else:
            return gpd.GeoDataFrame({'id': i, 'geometry': b}
                                    for i, b in enumerate(self))


class BBox(Polygon):
    """Sub-class of Polygon describing a BBox with a given id"""

    def __init__(self, bounds, bbox_id=None):
        self.left, self.bottom, self.right, self.top = bounds
        super().__init__([(self.left, self.top), (self.right, self.top),
                          (self.right, self.bottom), (self.left, self.bottom),
                          (self.left, self.top)])

        self.id = bbox_id

        self.width = self.right - self.left
        self.height = self.top - self.bottom

    def intersect(self, geo_df, buffer=0):
        """
        Returns list of overlapping sentinel 2 tiles given
        with s2grid GeoDataFrame
        """

        def inter_f(x): return x.geometry.buffer(buffer).intersects(self)
        flag = geo_df.apply(inter_f, axis=1)
        return geo_df[flag]

    def plot(self, *args, **kwargs):
        kwargs['facecolor'] = kwargs.get('facecolor', 'none')
        gpd.GeoDataFrame([{'id': 0,
                           'geometry': self}]).plot(*args, **kwargs)


def tile_to_epsg(tile):
    row = tile[2]
    zone = tile[:2]

    if row in 'CDEFGHJKLM':
        hemisphere = 'S'
    elif row in 'NPQRSTUVWX':
        hemisphere = 'N'
    else:
        raise ValueError(f"Unrecognized UTM zone '{zone}'.")

    utm = zone + hemisphere
    return utm_to_epsg(utm)


def utm_to_epsg(utm):
    utm = utm.upper()
    sud = 1 if utm[-1] == 'S' else 0
    zone = int(utm[:-1])
    epsg = 32600 + sud * 100 + zone
    return epsg


def get_tile_blocks(tile, s2grid=None, resolution=10):

    if s2grid is None:
        s2grid = satio.layers.load('s2grid')

    width = heigth = {10: 10980,
                      20: 5490,
                      60: 1830}[resolution]

    epsg = tile_to_epsg(tile)
    tile_bounds = s2grid[s2grid.tile == tile].to_crs(
        epsg=epsg).bounds.values[0].round().tolist()
    tile_transform = rasterio.transform.from_bounds(
        *tile_bounds, width, heigth)

    windows_tuples = load_s2tile_windows(resolution)

    polygons = []
    for t in windows_tuples:
        w = t[1]
        xmin, ymax = tile_transform * (w.col_off, w.row_off)
        xmax, ymin = tile_transform * \
            (w.col_off + w.width, w.row_off + w.height)

        polygons.append(Polygon.from_bounds(xmin, ymin, xmax, ymax))

    return gpd.GeoSeries(polygons, crs=CRS.from_epsg(epsg))


def get_blocks_gdf(tiles, s2grid, resolution=10):
    tiles_blocks = []
    for t in tiles:
        tblocks = get_tile_blocks(t, s2grid, resolution=resolution)
        tblocks_ll = tblocks.to_crs(epsg=4326)
        tiles_blocks += [{'tile': t,
                          'bounds': tuple(np.round(
                              np.array(b.bounds) / resolution).astype(int)
                              * resolution),
                          'geometry': tblocks_ll.iloc[i],
                          'area': b.area,
                          'epsg': tile_to_epsg(t),
                          'block_id': i}
                         for i, b in enumerate(tblocks)]
    df = gpd.GeoDataFrame(tiles_blocks, crs=CRS.from_epsg(4326))

    return df


def filter_land_blocks(blocks,
                       landsea,
                       s2grid,
                       landsea_buffer=0.03,
                       max_workers=4):

    s2tiles = s2grid[s2grid.tile.isin(blocks.tile.unique())]

    landsea_tiles = landsea.intersection(s2tiles.unary_union)
    landsea_geom = landsea_tiles.iloc[0].buffer(landsea_buffer)

    def f(row):
        if row.geometry.intersects(landsea_geom):
            return row.Index
        else:
            return None

    with ThreadPoolExecutor(max_workers=max_workers) as ex:
        valid_ids = list(ex.map(f, blocks.itertuples()))

    valid_ids = [r for r in valid_ids if r is not None]

    blocks = blocks.loc[valid_ids]

    return blocks


def filter_regions_blocks(blocks,
                          regions,
                          s2grid,
                          regions_buffer=0.03,
                          max_workers=4):

    s2tiles = s2grid[s2grid.tile.isin(blocks.tile.unique())]

    regions_tiles = regions.intersection(s2tiles.unary_union)
    regions_geom = regions_tiles.unary_union.buffer(regions_buffer)

    def f(row):
        if row.geometry.intersects(regions_geom):
            return row.Index
        else:
            return None

    with ThreadPoolExecutor(max_workers=max_workers) as ex:
        valid_ids = list(ex.map(f, blocks.itertuples()))

    valid_ids = [r for r in valid_ids if r is not None]

    blocks = blocks.loc[valid_ids]

    return blocks


def split_rectangle(geom):
    """
    Split rectangle in 4 rectangles
    """
    xmin, ymin, xmax, ymax = geom.bounds

    xmid = xmin + (xmax - xmin) / 2
    ymid = ymin + (ymax - ymin) / 2

    return [Polygon.from_bounds(xmin, ymin, xmid, ymid),
            Polygon.from_bounds(xmin, ymid, xmid, ymax),
            Polygon.from_bounds(xmid, ymin, xmax, ymid),
            Polygon.from_bounds(xmid, ymid, xmax, ymax)]


def get_tlocs_areas(tlocs, base_geoms=None, max_locs=300, final_geoms=[]):
    """
    Generate a grid of boxes by recursively splitting rectangles in 4 until
    every rectangle contains at most `max_locs` training points from `tlocs`
    """
    if base_geoms is None:
        base_geoms = [Polygon.from_bounds(-180, -90, 180, 90)]

    for g in base_geoms:
        n_locs = tlocs[tlocs.intersects(g)].shape[0]
        if n_locs <= max_locs:
            final_geoms.append(g)
        else:
            final_geoms = get_tlocs_areas(tlocs,
                                          base_geoms=split_rectangle(g),
                                          final_geoms=final_geoms)

    tlocs_areas = gpd.GeoSeries(final_geoms, crs=tlocs.crs)

    return tlocs_areas


def buffer_bounds(bounds, buffer):
    bounds = np.array(bounds)
    bounds += np.array([-buffer, -buffer, buffer, buffer])
    return bounds.tolist()


def clip_to_global_bbox(df):
    bbox = gpd.GeoSeries([Polygon([(-180, 90), (180, 90),
                                   (180, -90), (-180, -90)])])
    dfbbox = gpd.GeoDataFrame({'geometry': bbox,
                               'gbbox': 0}, crs=CRS.from_epsg(4326))
    dfint = gpd.tools.overlay(df, dfbbox)
    return dfint


def fiona_transform(df, dst_crs=None, dst_epsg=None):
    if dst_epsg is not None:
        dst_crs = CRS.from_epsg(dst_epsg)
    src_crs = df.crs

    def f(x): return fiona_transformer(src_crs, dst_crs, x)

    tdf = df.set_geometry(df.geometry.apply(f))
    tdf.crs = dst_crs
    return tdf


def fiona_transformer(src_crs, dst_crs, geom):
    fi_geom = transform_geom(src_crs=src_crs,
                             dst_crs=dst_crs,
                             geom=mapping(geom),
                             antimeridian_cutting=True)
    return shape(fi_geom)
