import numpy as np


NODATA_VALUE = -2**15

RSI_META_S2 = {
    'ndvi': {
        'bands': ['B08', 'B04'],
        'range': [-1, 1],
        'clamp_range': False,
        'scale': 2000,
        'native_res': 10},

    'ndmi': {'bands': ['B08', 'B11'],
             'range': [-1, 1],
             'clamp_range': False,
             'scale': 10000,
             'native_res': 20},

    'nbr': {'bands': ['B08', 'B12'],
            'range': [-1, 1],
            'clamp_range': False,
            'scale': 10000,
            'native_res': 20},

    'nbr2': {'bands': ['B11', 'B12'],
             'range': [-3, 3],
             'clamp_range': False,
             'scale': 10000,
             'native_res': 20},

    'evi': {'bands': ['B08', 'B04', 'B02'],
            'range': [-3, 3],
            'clamp_range': False,
            'scale': 10000,
            'native_res': 20},

    'evi2': {'bands': ['B08', 'B04'],
             'range': [-3, 3],
             'clamp_range': False,
             'scale': 10000,
             'native_res': 20},

    'savi': {'bands': ['B08', 'B04'],
             'range': [-3, 3],
             'clamp_range': False,
             'scale': 10000,
             'native_res': 10},

    'sipi': {'bands': ['B08', 'B01', 'B04'],
             'range': [-10, 10],
             'clamp_range': False,
             'scale': 1000,
             'native_res': 60},

    'hsvh': {'bands': ['B04', 'B03', 'B02'],
             'range': [0, 1],
             'clamp_range': False,
             'scale': 18000,
             'native_res': 10},

    'hsvv': {'bands': ['B04', 'B03', 'B02'],
             'range': [0, 1],
             'clamp_range': False,
             'scale': 2000,
             'native_res': 10},

    'hsv': {'bands': ['B04', 'B03', 'B02'],
            'range': [0, 1],
            'clamp_range': False,
            'scale': 2000,
            'native_res': 10,
            'output_bands': ['hsvh', 'hsvv']},

    'rep': {'bands': ['B04', 'B07', 'B05', 'B06'],
            'range': [600, 800],
            'clamp_range': False,
            'scale': 10,
            'native_res': 20},

    'anir': {'bands': ['B04', 'B08', 'B11'],
             'range': [0, 1],
             'clamp_range': False,
             'scale': 1,
             'native_res': 20},

    'nirv': {'bands': ['B08', 'B04'],
             'range': [-1, 1],
             'clamp_range': False,
             'scale': 10000,
             'native_res': 10},

    'auc': {'bands': ['B02', 'B04', 'B08', 'B11'],
            'range': [0, 1],
            'clamp_range': False,
            'scale': 20000,
            'native_res': 20},

    'nauc': {'bands': ['B02', 'B04', 'B08', 'B11'],
             'range': [0, 1],
             'clamp_range': False,
             'scale': 20000,
             'native_res': 20},

    # ndwi (mcFeeters)
    'ndwi': {'bands': ['B03', 'B08'],
             'range': [-1, 1],
             'clamp_range': False,
             'scale': 1,
             'native_res': 10},

    # normalized difference greenness index
    'ndgi': {'bands': ['B03', 'B04'],
             'range': [0, 1],
             'clamp_range': False,
             'scale': 1,
             'native_res': 10},

    # brightness (as defined in sen2agri)
    'brightness': {'bands': ['B03', 'B04', 'B08', 'B11'],
                   'range': [0, 1],
                   'clamp_range': False,
                   'scale': 1,
                   'native_res': 20},

    # series of normalized difference rede edge indices
    'ndre1': {'bands': ['B08', 'B05'],
              'range': [0, 1],
              'clamp_range': False,
              'scale': 1,
              'native_res': 20},

    'ndre2': {'bands': ['B08', 'B06'],
              'range': [0, 1],
              'clamp_range': False,
              'scale': 1,
              'native_res': 20},

    'ndre3': {'bands': ['B08', 'B07'],
              'range': [0, 1],
              'clamp_range': False,
              'scale': 1,
              'native_res': 20},

    'ndre4': {'bands': ['B06', 'B05'],
              'range': [0, 1],
              'clamp_range': False,
              'scale': 1,
              'native_res': 20},

    'ndre5': {'bands': ['B07', 'B05'],
              'range': [0, 1],
              'clamp_range': False,
              'scale': 1,
              'native_res': 20}}

RSI_META_S1 = {'vh_vv': {
    'bands': ['VH', 'VV'],
    'range': [0, 1],
    'clamp_range': False,
    'scale': 1,
    'native_res': 20}}

RSI_META = {'S2': RSI_META_S2,
            'S1': RSI_META_S1}

WL_B02, FWHM_B02 = 0.4924, 0.098
WL_B04, FWHM_B04 = 0.6646, (0.039 + 0.038) / 2
WL_B08, FWHM_B08 = 0.8328, (0.145 + 0.133) / 2
WL_B11, FWHM_B11 = 1.610, (0.143 + 0.141) / 2

# supported_vis = VI_META.keys()


def get_rsi_function(rsi_name, meta=None):
    """
    Derive RSI function either from its name or from meta
    The rsi_name argument suffices for indices defined within satio
    If a custom RSI is required, the function should be defined in 
    meta, as a callable under the 'func' key.

    :param rsi_name: string
    :param meta: optional dictionary containing a 'func' key

    """
    if meta is not None and 'func' in meta.keys():
        f = meta['func']
    else:
        if rsi_name in ['ndvi', 'ndmi', 'nbr', 'nbr2', 'ndwi', 'ndgi', 'ndre1', 'ndre2', 'ndre3', 'ndre4', 'ndre5']:
            f = norm_diff
        else:
            f = eval(rsi_name)
    return f


def norm_diff(arr1, arr2):
    """Returns the normalized difference of two bands"""
    return (arr1 - arr2) / (arr1 + arr2)


def evi(B08, B04, B02):
    return 2.5 * (B08 - B04) / (B08 + 6.0 * B04 - 7.5 * B02 + 1.0)


def evi2(B08, B04):
    return 2.5 * (B08 - B04) / (B08 + 2.4 * B04 + 1.0)


def savi(B08, B04):
    L = 0.428
    return (B08 - B04) / (B08 + B04 + L) * (1.0 + L)


def sipi(B08, B01, B04):
    return (B08 - B01) / (B08 - B04)


def hsv(B04, B03, B02):
    """Returns hsv 3d array from RGB bands"""
    nodata = np.isnan(B04)
    h, v = get_hsv_hue_value(B04, B03, B02)
    h[nodata] = np.nan
    v[nodata] = np.nan
    return np.array([h, v])


def hsvh(B04, B03, B02):
    hv = hsv(B04, B03, B02)
    return hv[0]


def hsvv(B04, B03, B02):
    hv = hsv(B04, B03, B02)
    return hv[1]


def rep(B04, B07, B05, B06):
    return 700 + 40 * ((((B04 + B07) / 2) - B05) / (B06 - B05))


def anir(B04, B08, B11):
    a = np.sqrt(np.square(WL_B08 - WL_B04) + np.square(B08 - B04))
    b = np.sqrt(np.square(WL_B11 - WL_B08) + np.square(B11 - B08))
    c = np.sqrt(np.square(WL_B11 - WL_B04) + np.square(B11 - B04))

    # calculate angle with NIR as reference (ANIR)
    site_length = (np.square(a) + np.square(b) - np.square(c)) / (2 * a * b)
    site_length[site_length < -1] = -1
    site_length[site_length > 1] = 1

    return 1. / np.pi * np.arccos(site_length)


def nirv(B08, B04):
    return ((B08 - B04 / B08 + B04) - 0.08) * B08


def auc(B02, B04, B08, B11):
    return B02 * FWHM_B02 + B04 * FWHM_B04 + B08 * FWHM_B08 + B11 * FWHM_B11


def nauc(B02, B04, B08, B11):
    min_ref = np.fmin(np.fmin(B02, B04), np.fmin(B08, B11))
    return ((B02 - min_ref) * FWHM_B02
            + (B04 - min_ref) * FWHM_B04
            + (B08 - min_ref) * FWHM_B08
            + (B11 - min_ref) * FWHM_B11)


def get_hsv_timeseries(r, g, b):

    h, s, v = np.zeros(r.shape), np.zeros(r.shape), np.zeros(r.shape)

    rgb = np.array([r, g, b])
    mx = np.nanmax(rgb, 0)
    mn = np.nanmin(rgb, 0)

    diff = mx - mn

    h[mx == mn] = 0
    h[mx == r] = ((60 * ((g - b) / diff) + 360) % 360)[mx == r]
    h[mx == g] = ((60 * ((b - r) / diff) + 360) % 360)[mx == g]
    h[mx == b] = ((60 * ((r - g) / diff) + 360) % 360)[mx == b]
    h = h / 360

    s = diff / mx
    s[mx == 0] = 0

    v = mx

    return h, s, v


def get_hsv_hue_value(r, g, b):
    h = np.zeros(r.shape, dtype=np.float32)

    rgb = np.array([r, g, b])
    mx = np.nanmax(rgb, 0)
    mn = np.nanmin(rgb, 0)

    diff = mx - mn

    h[mx == mn] = 0
    with np.errstate(divide='ignore', invalid='ignore'):  # type: ignore
        h[mx == r] = ((60 * ((g - b) / diff) + 360) % 360)[mx == r]
        h[mx == g] = ((60 * ((b - r) / diff) + 360) % 360)[mx == g]
        h[mx == b] = ((60 * ((r - g) / diff) + 360) % 360)[mx == b]

    h = h / 360
    v = mx

    return h, v


def brightness(B03, B04, B08, B11):
    return np.sqrt(np.power(B03, 2) + np.power(B04, 2) + np.power(B08, 2) + np.power(B11, 2))


def vh_vv(VH, VV):
    return VH / VV
