#from __future__ import annotations
from typing import Dict, List, Callable
import dateutil.parser

import scipy
import pandas as pd
import numpy as np
from loguru import logger

from satio.features import Features, compute_chunked_features, FEATURES_META
from satio.rsindices import get_rsi_function, RSI_META
from satio.utils.composite import calculate_moving_composite
from satio.utils.resample import block_downsample_band3d, upsample, imresize
from satio.utils.interpolation import interpolate_4d


def load_timeseries(collection, *bands, mask_and_scale=False) -> 'Timeseries':
    """
    Returns a Timeseries from a filtered collection
    """
    bands = list(bands)

    if mask_and_scale:
        data = collection.load(bands=bands, mask_and_scale=mask_and_scale)
    else:
        # Disk Collection doesn't have mask_and_scale kwarg, avoid raising
        data = collection.load(bands=bands)

    arr = [data[b].values for b in bands]

    if not all([d.shape == arr[0].shape for d in arr]):
        raise ValueError(f"Bands {bands} do not have all the same "
                         "resolution.")

    timestamp = data[bands[0]].timestamp.values
    timestamp = np.array([dateutil.parser.parse(str(s)) for s in timestamp])

    arr = np.array(arr)

    # if arr.dtype == np.int16:
    #     logger.warning("Data is int16, forcing uint16")
    #     arr[arr < 0] = 0
    #     arr = arr.astype(np.uint16)
    
    attrs = {'sensor': collection.sensor}

    return Timeseries(arr, timestamp, bands, attrs)


class Timeseries:

    def __init__(self, data: np.ndarray, timestamps: List, bands: List, attrs: dict = dict()):
        if data.ndim != 4:
            raise ValueError("Invalid data input shape")
        if data.shape[0] != len(bands):
            raise ValueError(
                "Inconsistent data dimension 0 and number of bands")
        if data.shape[1] != len(timestamps):
            raise ValueError(
                "Inconsistent data dimension 1 and number of timestamps")

        self.data = data
        self.timestamps = timestamps
        self.bands = list(bands)
        self.attrs = attrs

    def _clone(self, data=None, timestamps=None, bands=None, attrs: dict = None) -> 'Timeseries':
        data = self.data if data is None else data
        timestamps = self.timestamps if timestamps is None else timestamps
        bands = self.bands if bands is None else bands
        attrs = self.attrs if attrs is None else attrs

        return Timeseries(data, timestamps, bands, attrs)

    def composite(self,
                  freq=7,
                  before=7,
                  after=7,
                  start='20190101',
                  end='20200101',
                  percentile=50) -> 'Timeseries':
        """
        Return a moving median composite timeseries. The `freq`, `before`
        `after` parameters are in days.
        """
        new_arr, new_timestamps = calculate_moving_composite(
            self.data,
            self.timestamps,
            start=start,
            end=end,
            freq=freq,
            before=before,
            after=after,
            percent=percentile)

        return self._clone(data=new_arr, timestamps=new_timestamps)  # type: ignore  # NOQA

    def interpolate(self, inplace=False) -> 'Timeseries':
        """
        Interpolate missing values (np.nan for float32 arrays or 0 for uint16)
        Other dtypes are not supported.
        """
        if inplace:
            data = self.data
        else:
            data = self.data.copy()

        """
        there is a bug in the interpolate function. For a full timeseries
        of 0 the first pixel can be different from zero after the interpolation
        this monkey patch ensures that all zeros go out as all zeros
        """

        if data.dtype == np.uint16:
            pre_interp_mask = (data[0] > 0).sum(axis=0) > 0
        elif data.dtype == np.float32:
            pre_interp_mask = (~np.isnan(data[0])).sum(axis=0) > 0
        else:
            raise TypeError(f"dtype {data.dtype} not supported.")

        new_data = interpolate_4d(data)

        new_data[:, :, ~pre_interp_mask] = (0 if data.dtype == np.uint16
                                            else np.nan)

        return self._clone(data=new_data)

    def mask(self, mask, nodata_value=None, drop_nodata=True) -> 'Timeseries':
        """
        Mask data values with the given 2d or 3d mask.
        Values False or 0 in the mask will be set to nodata_value.
        Timestamps with no valid data will be dropped by the Timeseries.
        """
        if nodata_value is None:
            if self.data.dtype == np.uint16:
                nodata_value = 0
            elif self.data.dtype == np.float32:
                nodata_value = np.nan
            else:
                raise ValueError(f'dtype {self.data.dtype} not supported.')

        if ((self.data.dtype == np.uint16) &
                ((nodata_value < 0) | (nodata_value >= 2**16))):
            raise ValueError("`nodata_value` for uint16 timeseries "
                             "should be between 0 and 65535")

        mask = np.broadcast_to(mask, self.data.shape)

        masked_data = self.data.copy()
        masked_data[~mask] = nodata_value

        new = self._clone(data=masked_data)

        if drop_nodata:
            new = new.drop_nodata()

        return new

    def drop_nodata(self, ref_band=0):
        """
        Return new timeseries without nodata timestamps from ref_band
        """
        if self.data.dtype == np.uint16:
            nodata_value = 0
        elif self.data.dtype == np.float32:
            nodata_value = np.nan
        else:
            raise ValueError(f'dtype {self.data.dtype} not supported.')
        
        s = self.data[ref_band].sum(axis=(1, 2))
        nodata_ids = np.where(s == nodata_value)[0]
        return self.drop_ids(nodata_ids)

    def drop_ids(self, ids) -> 'Timeseries':
        """
        Return new timeseries withouth the given timestamps
        ids.
        """
        valid_ids = [i for i in range(self.data.shape[1])
                     if i not in ids]
        new_data = self.data[:, valid_ids, :, :]
        new_timestamps = self.timestamps[valid_ids]

        return self._clone(data=new_data, timestamps=new_timestamps)

    def select_bands(self, bands) -> 'Timeseries':
        """
        Returns timeseries from the selected bands
        """
        bands_ids = []
        for b in bands:
            if b not in self.bands:
                raise ValueError(f"Band {b} not available.")
        bands_ids = [self.bands.index(b) for b in bands]

        new_data = self.data[bands_ids, ...]

        return self._clone(data=new_data, bands=bands)

    def combine_bands(self,
                      func: Callable,
                      bands: List,
                      output_bands: List,
                      bands_scaling: float = 10000) -> 'Timeseries':

        new_ts = self.select_bands(bands)

        if bands_scaling is None:
            bands_scaling = 1

        if (new_ts.data.dtype == np.uint16) & (bands_scaling == 1):
            logger.warning("Combining bands of dtype uint16 "
                           "without scaling values to reflectance.")

        # convert to float before computing rsis
        new_ts.data = new_ts.data.astype(np.float32) / bands_scaling

        new_data = func(*[np.squeeze(new_ts[b].data) for b in bands])

        if new_data.ndim == 3:
            new_data = np.expand_dims(new_data, axis=0)

        return self._clone(data=new_data, bands=output_bands)

    def merge(self, *others) -> 'Timeseries':
        """
        Returns a new instance merging data and names of current feature with
        multiple timeseries.
        """
        feat = self
        for ot in others:
            feat = feat._merge(ot)

        return feat

    def _merge(self, other) -> 'Timeseries':
        """
        Merge two timeseries with same timestamps and spatial
        extension. If the same band is present in `other` and `self`
        the one in `other` will be discarded
        """
        if any(self.timestamps != other.timestamps):
            raise ValueError("Cannot merge two timeseries "
                             "with different time vectors.")

        if (self.data.shape[-2:] != other.data.shape[-2:]):
            raise ValueError("Cannot merge two timeseries with "
                             "different canvas shape")

        new_other_bands = [b for b in other.bands if b not in self.bands]

        if len(new_other_bands) == 0:
            return self

        new_other = other.select_bands(new_other_bands)
        new_data = np.concatenate([self.data, new_other.data], axis=0)
        new_bands = self.bands + new_other_bands

        return self._clone(data=new_data, bands=new_bands)

    def downsample(self) -> 'Timeseries':
        """
        Downsamples timeseries to half resolution in the space dimensions.
        Only works for uint16. For other types, see `imresize` method.
        """
        new_data = np.array([block_downsample_band3d(np.squeeze(self[b].data))
                             for b in self.bands])
        return self._clone(data=new_data)

    def upsample(self) -> 'Timeseries':
        """
        Upsamples timeseries to double the resolution in the space dimensions.
        Using a nearest neighbor method.
        For other resampling methods/scalings, see `imresize` method.
        """
        new_data = np.array([upsample(np.squeeze(self[b].data))
                             for b in self.bands])
        return self._clone(data=new_data)

    def imresize(self, scaling, order=1):
        """
        Resample the timeseries spatially. When upsampling uses
        `skimage.transform.rescale` with the given order (method) given.
        When downsampling either uses `block_resize` when possible or
        `rescale` if the pixels dimensions require it.
        """
        new_data = []
        for b in self.bands:
            band_data = np.squeeze(self[b].data)
            new_band = []
            for i in range(band_data.shape[0]):
                new_band.append(imresize(band_data[i], scaling, order))

            new_data.append(new_band)

        new_data = np.array(new_data)

        return self._clone(data=new_data)

    def resample_time(self, periods: int) -> 'Timeseries':
        """
        Resample timeseries along the time dimensions using
        `scipy.signal.resample`.
        Timestamps should be regularly spaced as this method
        uses an FFT transform.
        """
        new_data = scipy.signal.resample(self.data, periods, axis=1)
        new_timestamps = pd.date_range(start=self.timestamps[0],
                                       end=self.timestamps[-1],
                                       periods=periods).to_pydatetime()

        return self._clone(data=new_data, timestamps=new_timestamps)

    def compute_band_features(self,
                              func: Callable,
                              band: str,
                              features_names: List,
                              chunk_size: int = None,
                              **kwargs):
        """
        `func` should be a function that acts on a numpy array of
        shape [time, y, x] along the time dimension, returning an array of
        shape [len(features_names), y, x].

        The function is called as `func(arr, **kwargs)`
        """
        data = np.squeeze(self.select_bands([band]).data)

        if chunk_size is None:
            features = func(data, **kwargs)
        else:
            n_features = len(features_names)
            features = compute_chunked_features(data,
                                                func,
                                                n_features,
                                                chunk_size=chunk_size,
                                                **kwargs)

        return Features(features, features_names)

    def features_from_dict(self,
                           resolution,
                           features_meta=FEATURES_META,
                           chunk_size=None):

        features = []
        for k, v in features_meta.items():
            f = v['function']
            params = v.get('parameters', {})
            feature_bands = v.get('bands', self.bands)
            bands = [b for b in feature_bands if b in self.bands]
            for b in bands:
                features_names = [
                    f'{b}-{vn}-{resolution}m' for vn in v['names']]
                features.append(
                    self.compute_band_features(f, b, features_names,
                                               chunk_size=chunk_size,
                                               **params))

        if len(features) > 1:
            features = Features.from_features(*features)
        elif len(features) == 1:
            features = features[0]

        if type(features) is list:
            features = None
            
        return features

    def compute_rsis(self,
                    *rsi_names,
                    rsi_meta: Dict = {},
                    bands_scaling: int = 10000) -> 'Timeseries':
        """
        Returns a Timeseries from the combination of bands.

        `bands_scaling` is set to 10000 by default to scale S2 bands to
        reflectance
        """
        rsi_meta = {**RSI_META[self.attrs['sensor']], **rsi_meta}            

        rsis_ts = None

        for rsi_name in rsi_names:
            ts = self._compute_rsi(rsi_name,
                                  rsi_meta=rsi_meta,
                                  bands_scaling=bands_scaling)

            if rsis_ts is None:
                rsis_ts = ts
            else:
                rsis_ts = rsis_ts.merge(ts)

        return rsis_ts

    def _compute_rsi(self,
                    rsi_name: str,
                    rsi_meta: Dict = None,
                    bands_scaling: int = 10000):
        """
        Returns a Timeseries from the combination of bands
        """
        meta = rsi_meta[rsi_name]
        
        func = get_rsi_function(rsi_name, meta=meta)
        
        rsi_bands = meta['bands']

        output_bands = meta.get('output_bands', [rsi_name])
        clamp = meta['clamp_range']
        range_max, range_min = meta['range']

        rsi_ts = self.combine_bands(func,
                                   rsi_bands,
                                   output_bands=output_bands,
                                   bands_scaling=bands_scaling)

        if clamp:
            rsi_ts.data[rsi_ts.data < range_min] = range_min
            rsi_ts.data[rsi_ts.data > range_max] = range_max

        if (~np.isfinite(rsi_ts.data)).sum() > 0:
            rsi_ts = rsi_ts.interpolate()

        return rsi_ts

    def show(self,
             mask=None,
             vmin=None,
             vmax=None,
             rgb_indices=[0, 1, 2],
             **figure_options):

        from satio.utils.visualizer import Visualizer
        data = self.data
        bands = self.bands

        if data.shape[0] < 3:
            if data.shape[0] == 2:
                logger.warning("Cannot use Visualizer with 2 bands."
                               " Selecting first band.")
            data = data[0]
            bands = [bands[0]]

        return Visualizer(data,
                          self.timestamps,
                          bands_labels=bands,
                          mask=mask,
                          vmin=vmin,
                          vmax=vmax,
                          rgb_indices=rgb_indices,
                          **figure_options)

    @classmethod
    def from_timeseries(cls, timeseries):
        return timeseries[0].merge(*timeseries[1:])

    def __getitem__(self, band) -> 'Timeseries':
        return self.select_bands([band])

    def __repr__(self):
        return (f"<Timeseries: {self.data.shape} - bands: {self.bands} "
                f"- s: {self.timestamps[0]} - e: {self.timestamps[-1]} "
                f"dtype: {self.data.dtype}>")
