from datetime import timedelta

import numpy as np
import pandas as pd


def interval_flag(time_vector, date, before, after):
    """
    Returns a boolean array True where the dates in time_vector fall in an
    interval between data - before and date + after

    Parameters
    ----------
    time_vector : datetime/np.datetime64 array
        time_vector of the source.
    date : datetime/np.datetime64
        target date from which we want to get the neighboring dates from
        time_vector
    """
    return ((time_vector >= date + timedelta(days=-before))
            & (time_vector <= date + timedelta(days=after)))


def calculate_moving_composite(arrs,
                               time_vector,
                               start='20171001',
                               end='20190401',
                               freq=7,
                               before=7,
                               after=7,
                               percent=50):
    """
    Calculate moving median composite of an hyper cube with shape [bands, time,
    rows, cols]

    Parameters
    ----------
    arrs : Numpy 4d array [bands, time, rows, cols]

    start : str or datetime
        start date for the new composites dates
    end : str or datetime
        end date for the new composites dates
    freq : int
        days interval for new dates array
    before: int
        window before date for the compositing
    after: int
        window after date for the compositing
    percent: int
        between 0 and 100. Usually 50 for the median compositing

    Return
    ----------
    Tuple of time_vector and composite 4d array
    """
    if arrs.dtype == np.uint16:
        from satio.utils.nonzeropercentile import (hyper_nonzeropercentile,
                                                   nonzeropercentile)

        hyper_n_percentile = hyper_nonzeropercentile
        n_percentile = nonzeropercentile

    elif arrs.dtype == np.float32:
        from satio.utils.nanpercentile import (hyper_nanpercentile,
                                               nanpercentile)

        hyper_n_percentile = hyper_nanpercentile
        n_percentile = nanpercentile
    else:
        raise TypeError(f'{arrs.dtype} not supported for numba nan/nonzero '
                        'percentiles calculation. Should be uint16 or float32')

    date_range = pd.date_range(start=start, end=end, freq='{}D'.format(freq))

    if len(arrs.shape) == 4:
        comp = np.zeros((arrs.shape[0], len(date_range),
                         arrs.shape[2], arrs.shape[3]), dtype=arrs.dtype)

        for i, d in enumerate(date_range):
            idxs = interval_flag(pd.to_datetime(
                time_vector), d, before=before, after=after)
            comp[:, i, ...] = hyper_n_percentile(arrs[:, idxs, ...],
                                                    percent)
                
    elif len(arrs.shape) == 3:
        comp = np.zeros((len(date_range),
                         arrs.shape[1], arrs.shape[2]), dtype=arrs.dtype)

        for i, d in enumerate(date_range):
            idxs = interval_flag(pd.to_datetime(
                time_vector), d, before=before, after=after)
            comp[i, ...] = n_percentile(arrs[idxs, ...], percent)

    date_range = date_range.to_pydatetime()

    return comp, date_range  # type: ignore
